# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import overload

import logging

from gi.repository import Gtk
from nbxmpp.errors import StanzaError
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import ged
from gajim.common.events import AccountConnected
from gajim.common.events import AccountDisconnected
from gajim.common.helpers import to_user_string
from gajim.common.i18n import _
from gajim.common.util.decorators import event_filter

from gajim.gtk.assistant import Assistant
from gajim.gtk.assistant import AssistantErrorPage
from gajim.gtk.assistant import AssistantPage
from gajim.gtk.assistant import AssistantSuccessPage

log = logging.getLogger("gajim.gtk.remove_account")


class RemoveAccount(Assistant):
    def __init__(self, account: str) -> None:
        Assistant.__init__(self)

        self.account = account

        try:
            self._client = app.get_client(account)
        except KeyError:
            self._client = None

        self._destroyed = False
        self._account_removed = False

        self.add_button("remove", _("Remove"), "destructive-action")
        self.add_button("close", _("Close"))
        self.add_button("back", _("Back"))

        self.add_pages(
            {
                "remove_choice": RemoveChoice(account),
                "error": Error(),
                "success": Success(),
            }
        )

        progress = self.add_default_page("progress")
        progress.set_title(_("Removing Account..."))
        progress.set_text(_("Trying to remove account..."))

        self._connect(self, "button-clicked", self._on_button_clicked)

        self.register_events(
            [
                ("account-connected", ged.POSTGUI, self._on_account_connected),
                ("account-disconnected", ged.POSTGUI, self._on_account_disconnected),
            ]
        )

        self._set_remove_from_server_checkbox()

    @overload
    def get_page(self, name: Literal["remove_choice"]) -> RemoveChoice: ...

    @overload
    def get_page(self, name: Literal["error"]) -> Error: ...

    @overload
    def get_page(self, name: Literal["success"]) -> Success: ...

    def get_page(self, name: str) -> AssistantPage:
        return self._pages[name]

    @event_filter(["account"])
    def _on_account_connected(self, _event: AccountConnected) -> None:
        self._client = app.get_client(self.account)
        self._set_remove_from_server_checkbox()

    @event_filter(["account"])
    def _on_account_disconnected(self, _event: AccountDisconnected) -> None:
        self._set_remove_from_server_checkbox()

        if self._account_removed:
            self.show_page("success")
            app.app.remove_account(self.account)

    def _set_remove_from_server_checkbox(self) -> None:
        enabled = self._client is not None and self._client.state.is_available
        self.get_page("remove_choice").set_remove_from_server(enabled)

    def _on_button_clicked(self, _assistant: Assistant, button_name: str) -> None:
        page = self.get_current_page()
        if button_name == "remove":
            if page == "remove_choice":
                self.show_page("progress", Gtk.StackTransitionType.SLIDE_LEFT)
                self._on_remove()
            return

        if button_name == "back":
            if page == "error":
                self.show_page("remove_choice", Gtk.StackTransitionType.SLIDE_RIGHT)
            return

        if button_name == "close":
            self.close()

    def _on_remove(self, *args: Any) -> None:
        remove_choice_page = self.get_page("remove_choice")
        if remove_choice_page.remove_from_server:
            assert self._client is not None
            self._client.set_remove_account(True)
            self._client.get_module("Register").unregister(
                callback=self._on_remove_response
            )
            return

        if self._client is None or self._client.state.is_disconnected:
            app.app.remove_account(self.account)
            self.show_page("success")
            return

        self._client.disconnect(gracefully=True, reconnect=False)
        self._account_removed = True

    def _on_remove_response(self, task: Task) -> None:
        try:
            task.finish()
        except StanzaError as error:
            assert self._client is not None
            self._client.set_remove_account(False)

            error_text = to_user_string(error)
            self.get_page("error").set_text(error_text)
            self.show_page("error")
            return

        self._account_removed = True

    def _cleanup(self) -> None:
        Assistant._cleanup(self)
        self._destroyed = True


class RemoveChoice(AssistantPage):
    def __init__(self, account: str) -> None:
        AssistantPage.__init__(self)
        self.title = _("Remove Account")

        heading = Gtk.Label(
            label=_("Remove Account"),
            wrap=True,
            max_width_chars=30,
            halign=Gtk.Align.CENTER,
            justify=Gtk.Justification.CENTER,
        )
        heading.add_css_class("title-1")

        label = Gtk.Label(
            label=_("This will remove your account from Gajim."),
            wrap=True,
            max_width_chars=50,
            halign=Gtk.Align.CENTER,
            justify=Gtk.Justification.CENTER,
        )

        service = app.get_hostname_from_account(account)
        check_label = Gtk.Label(
            wrap=True,
            max_width_chars=50,
            halign=Gtk.Align.CENTER,
            justify=Gtk.Justification.CENTER,
            margin_top=40,
        )
        check_label.set_markup(
            _("Do you want to unregister your account on <b>%s</b> as well?") % service
        )

        self._server = Gtk.CheckButton.new_with_mnemonic(
            _("_Unregister account from service")
        )
        self._server.set_halign(Gtk.Align.CENTER)

        self.append(heading)
        self.append(label)
        self.append(check_label)
        self.append(self._server)

    @property
    def remove_from_server(self) -> bool:
        return self._server.get_active()

    def set_remove_from_server(self, enabled: bool) -> None:
        self._server.set_sensitive(enabled)
        if enabled:
            self._server.set_tooltip_text("")
        else:
            self._server.set_active(False)
            self._server.set_tooltip_text(_("Account has to be connected"))

    def get_visible_buttons(self) -> list[str]:
        return ["remove"]


class Error(AssistantErrorPage):
    def __init__(self) -> None:
        AssistantErrorPage.__init__(self)
        self.set_title(_("Account Removal Failed"))
        self.set_heading(_("Account Removal Failed"))

    def get_visible_buttons(self) -> list[str]:
        return ["back"]


class Success(AssistantSuccessPage):
    def __init__(self) -> None:
        AssistantSuccessPage.__init__(self)
        self.set_title(_("Account Removed"))
        self.set_heading(_("Account Removed"))
        self.set_text(_("Your account has has been removed successfully."))

    def get_visible_buttons(self) -> list[str]:
        return ["close"]



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing
from typing import TYPE_CHECKING

import sys
from concurrent.futures import Future
from functools import partial
from pathlib import Path

from gi.repository import GLib
from gi.repository import GObject
from gi.repository.Gdk import Paintable

from gajim.common import app
from gajim.common.multiprocess.animated_image_frames import extract_frames

try:
    from gi.repository import Gst
except Exception:
    if TYPE_CHECKING:
        from gi.repository import Gst

import logging

from gajim.gtk.util.classes import SignalManager

log = logging.getLogger("gajim.gtk.preview_animated_image_fallback_backend")


class AnimatedImageFallbackBackend(GObject.Object, SignalManager):
    __gtype_name__ = "AnimatedImageFallbackBackend"
    __gsignals__ = {
        "pipeline-changed": (GObject.SignalFlags.RUN_LAST, None, (bool,)),
        "playback-changed": (GObject.SignalFlags.RUN_LAST, None, (bool,)),
    }

    def __init__(self, orig_path: Path, max_loops: int = 3) -> None:
        super().__init__()
        SignalManager.__init__(self)

        self._orig_path = orig_path

        self._creating_pipeline = False
        self._pipeline_is_setup = False
        self._pipeline_setup_failed = False

        self._do_stop = True
        self._push_id = None

        self._paintable: Paintable | None = None
        self._pipeline: Gst.Element | None = None
        self._src: Gst.Element | None = None
        self._bus: Gst.Bus | None = None

        # Workaround a regression of Gst or Gtk on Windows,
        # wglShareLists() called by GstGL would fail due to ERROR_BUSY (0xAA)
        self._use_gl = sys.platform != "win32"

        self._buf: Gst.Buffer | None = None
        self._frames: list[tuple[bytes, int]] = []
        self._current_frame = 1

        self._loop_counter = 0
        self._max_loops = max_loops

    @property
    def pipeline_is_setup(self) -> bool:
        return self._pipeline_is_setup

    @property
    def pipeline_setup_failed(self) -> bool:
        return self._pipeline_setup_failed

    @property
    def paintable(self) -> Paintable | None:
        return self._paintable

    def cleanup(self) -> None:
        self._cleanup()
        self._disconnect_all()
        app.check_finalize(self)

    def is_playing(self) -> bool:
        if self._pipeline is None:
            return False
        _, state, _ = self._pipeline.get_state(timeout=100)
        return state == Gst.State.PLAYING

    def play(self) -> None:
        assert self._pipeline is not None
        self._do_stop = False
        self._pipeline.set_state(Gst.State.PLAYING)
        self._push_frame()

    def pause(self) -> None:
        if self._pipeline is None:
            self._pipeline_is_setup = False
            return

        _, state, _ = self._pipeline.get_state(timeout=100)
        if state == Gst.State.PLAYING:
            if self._push_id is not None:
                GLib.source_remove(self._push_id)
                self._push_id = None

        self._loop_counter = 0
        self._current_frame = 1
        self._do_stop = True
        self._pipeline.set_state(Gst.State.PAUSED)
        self._pipeline.send_event(Gst.Event.new_flush_start())
        self._pipeline.send_event(Gst.Event.new_flush_stop(True))
        self.emit("playback-changed", False)

    def setup_pipeline(self) -> None:
        try:
            self._setup_pipeline()
        except Exception:
            log.exception("Unable to play animated image")
            self._pipeline_setup_failed = True
            self.emit("pipeline-changed", False)

    def _setup_pipeline(self) -> None:
        if (
            self._creating_pipeline
            or self._pipeline_is_setup
            or self._pipeline_setup_failed
        ):
            return

        self._creating_pipeline = True
        self._current_frame = 1
        self._loop_counter = 0
        self._do_stop = True

        pipeline_parts = {
            "src": "appsrc name=src is-live=true format=time block=false leaky-type=2 do-timestamp=true",
            "dec": "webpdec name=webpdec bypass-filtering=false no-fancy-upsampling=true use-threads=true",
            "convert": "videoconvert name=videoconvert",
            "gl": "glupload name=glupload ! glcolorconvert" if self._use_gl else "",
            "queue": "queue name=queue",
            "sink": "gtk4paintablesink name=sink",
        }
        pipeline = " ! ".join(filter(None, pipeline_parts.values()))
        try:
            self._pipeline = typing.cast(Gst.Pipeline, Gst.parse_launch(pipeline))
        except Exception as e:
            self._pipeline_setup_failed = False
            self.emit("pipeline-changed", False)
            log.warning("Failed to setup pipeline: %s", e)
            return

        self._bus = self._pipeline.get_bus()
        assert self._bus is not None
        self._bus.add_signal_watch()
        self._connect(self._bus, "message::eos", self._on_eos)
        self._connect(self._bus, "message::state-changed", self._on_state_changed)

        sink: Gst.Element | None = self._pipeline.get_by_name("sink")
        assert sink is not None
        self._paintable = sink.get_property("paintable")
        self._src = self._pipeline.get_by_name("src")

        self._get_frames()

    def _on_state_changed(self, _bus: Gst.Bus, message: Gst.Message) -> None:
        assert self._pipeline is not None
        assert message is not None
        if message.src == self._pipeline:
            _, new, _ = message.parse_state_changed()
            if new == Gst.State.PLAYING:
                self.emit("playback-changed", True)

    def _get_frames(self) -> None:
        assert self._orig_path is not None
        try:
            future = app.process_pool.submit(extract_frames, self._orig_path)
            future.add_done_callback(
                partial(GLib.idle_add, self._extracting_frames_finished)
            )
        except Exception as error:
            log.exception("Extracting frames failed for: %s %s", self._orig_path, error)
            self._pipeline_setup_failed = True
            self.emit("pipeline-changed", False)

    def _extracting_frames_finished(
        self, future: Future[list[tuple[bytes, int]]]
    ) -> bool:
        try:
            self._frames = future.result()
        except Exception as error:
            log.exception("Extracting frames failed for: %s %s", self._orig_path, error)
            self._pipeline_setup_failed = True
            self.emit("pipeline-changed", False)
            return GLib.SOURCE_REMOVE

        self._pipeline_is_setup = True
        self.emit("pipeline-changed", True)
        return GLib.SOURCE_REMOVE

    def _push_frame(self):
        if self._do_stop:
            self._push_id = None
            return False

        frame, duration = self._frames[self._current_frame]
        buf = Gst.Buffer.new_wrapped(frame)
        assert buf is not None
        buf.duration = duration * Gst.MSECOND
        assert self._src is not None
        self._src.emit("push-buffer", buf)
        self._current_frame = (self._current_frame + 1) % len(self._frames)
        del buf
        self._push_id = GLib.timeout_add(duration, self._push_frame)
        if self._current_frame == 0:
            self._loop_counter = (self._loop_counter + 1) % self._max_loops
            if self._loop_counter == 0:
                self.pause()
        return False

    def _on_eos(self, _bus: Gst.Bus, _msg: Gst.Message) -> None:
        assert self._pipeline is not None
        self._pipeline.seek_simple(Gst.Format.TIME, Gst.SeekFlags.FLUSH, 0)

    def _cleanup(self) -> None:
        if self._push_id is not None:
            GLib.source_remove(self._push_id)

        if self._pipeline is not None:
            self._pipeline.set_state(Gst.State.NULL)
            self._pipeline.run_dispose()
            self._pipeline = None
        if self._src is not None:
            self._src.run_dispose()
        if self._bus is not None:
            self._bus.remove_signal_watch()
            self._bus = None

        self._paintable = None
        self._frames = []



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Literal

from nbxmpp.const import PresenceShow

from gajim.common import app
from gajim.common.const import PresenceShowExt
from gajim.common.const import SHOW_LIST
from gajim.common.const import SHOW_STRING
from gajim.common.const import SHOW_STRING_MNEMONIC

ShowSortOrder = {
    PresenceShow.CHAT: 1,
    PresenceShow.ONLINE: 1,
    PresenceShow.DND: 2,
    PresenceShow.AWAY: 3,
    PresenceShow.XA: 4,
    PresenceShowExt.OFFLINE: 5,
}


def get_client_status(account: str) -> str:
    client = app.get_client(account)
    if client.state.is_disconnected:
        return "offline"

    if (
        client.state.is_reconnect_scheduled
        or client.state.is_connecting
        or client.state.is_connected
    ):
        return "connecting"

    return client.status


def get_global_show() -> str:
    maxi = 0
    for client in app.get_clients():
        if not app.settings.get_account_setting(
            client.account, "sync_with_global_status"
        ):
            continue
        status = get_client_status(client.account)
        index = SHOW_LIST.index(status)
        maxi = max(index, maxi)

    return SHOW_LIST[maxi]


def get_global_status_message() -> str:
    maxi = 0
    status_message = ""
    for client in app.get_clients():
        if not app.settings.get_account_setting(
            client.account, "sync_with_global_status"
        ):
            continue
        index = SHOW_LIST.index(client.status)
        if index > maxi:
            maxi = index
            status_message = client.status_message
    return status_message


def statuses_unified() -> bool:
    """
    Test if all statuses are the same
    """
    reference = None
    for client in app.get_clients():
        account = client.account
        if not app.settings.get_account_setting(account, "sync_with_global_status"):
            continue

        if reference is None:
            reference = get_client_status(account)

        elif reference != get_client_status(account):
            return False
    return True


def get_uf_show(show: str, use_mnemonic: bool = False) -> str:
    if use_mnemonic:
        return SHOW_STRING_MNEMONIC[show]
    return SHOW_STRING[show]


def get_idle_status_message(
    state: Literal["away"] | Literal["xa"], status_message: str
) -> str:
    if state == "away":
        message = app.settings.get("autoaway_message")
        idle_time = app.settings.get("autoawaytime")
    else:
        message = app.settings.get("autoxa_message")
        idle_time = app.settings.get("autoxatime")

    if not message:
        return status_message

    message = message.replace("$S", "%(status)s")
    message = message.replace("$T", "%(time)s")

    return message % {"status": status_message, "time": idle_time}


def compare_show(
    show1: PresenceShow | PresenceShowExt,
    show2: PresenceShow | PresenceShowExt,
) -> int:
    val1 = ShowSortOrder[show1]
    val2 = ShowSortOrder[show2]
    if val1 == val2:
        return 0
    return -1 if val1 < val2 else 1



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging
from pathlib import Path

from gi.repository import Gio
from gi.repository import GLib

from gajim.common import configpaths
from gajim.common.helpers import write_file_async
from gajim.common.util.text import get_random_string

log = logging.getLogger('gajim.c.cert_store')


class CertificateStore:
    def __init__(self):
        self._path = configpaths.get('CERT_STORE')
        self._certs: list[Gio.TlsCertificate] = []

        self._load_certificates()

    def _get_random_path(self) -> Path:
        filename = get_random_string()
        path = self._path / filename
        if path.exists():
            return self._get_random_path()
        return path

    def _load_certificates(self) -> None:
        for path in self._path.iterdir():
            if path.is_dir():
                continue
            try:
                cert = Gio.TlsCertificate.new_from_file(str(path))
            except GLib.Error as error:
                log.warning("Can't load certificate: %s, %s", path, error)
                continue

            log.info('Loaded: %s', path.stem)
            self._certs.append(cert)

        log.info('%s Certificates loaded', len(self._certs))

    def get_certificates(self):
        return list(self._certs)

    def add_certificate(self, certificate: Gio.TlsCertificate) -> None:
        log.info('Add certificate to trust store')
        self._certs.append(certificate)
        pem = certificate.props.certificate_pem
        path = self._get_random_path()
        write_file_async(path,
                         pem.encode(),
                         self._on_certificate_write_finished,
                         path)

    def verify(self,
               certificate: Gio.TlsCertificate,
               tls_errors: set[Gio.TlsCertificateFlags]):

        if Gio.TlsCertificateFlags.UNKNOWN_CA in tls_errors:
            for trusted_certificate in self._certs:
                if trusted_certificate.is_same(certificate):
                    tls_errors.remove(Gio.TlsCertificateFlags.UNKNOWN_CA)
                    break

        return not tls_errors

    @staticmethod
    def _on_certificate_write_finished(_successful: bool,
                                       error: GLib.Error | None,
                                       path: Path):
        if error is not None:
            log.error("Can't store certificate: %s", error.message)
            return

        log.info('Certificate stored: %s', path)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import logging
import sys
from collections.abc import Callable

from gi.repository import Gio
from gi.repository import GLib

from gajim.common import app
from gajim.common.events import StyleChanged

log = logging.getLogger("gajim.c.dbus.system_style")


class SystemStyleListener:
    def __init__(self, callback: Callable[..., None]) -> None:
        self._prefer_dark: bool | None = None
        self._callback = callback

        if sys.platform in ("win32", "darwin"):
            return

        try:
            self.dbus_proxy = Gio.DBusProxy.new_for_bus_sync(
                Gio.BusType.SESSION,
                Gio.DBusProxyFlags.NONE,
                None,
                "org.freedesktop.portal.Desktop",
                "/org/freedesktop/portal/desktop",
                "org.freedesktop.portal.Settings",
                None,
            )
        except GLib.Error as error:
            log.info("Settings portal not found: %s", error)
            return

        self.dbus_proxy.connect("g-signal", self._signal_setting_changed)
        self.read_color_scheme()

    def read_color_scheme(self) -> None:
        try:
            result = self.dbus_proxy.call_sync(
                "Read",
                GLib.Variant("(ss)", ("org.freedesktop.appearance", "color-scheme")),
                Gio.DBusCallFlags.NO_AUTO_START,
                -1,
                None,
            )
            self._prefer_dark = result[0] == 1
        except GLib.Error as error:
            log.error("Couldn’t read the color-scheme setting: %s", error.message)
            return

    def _signal_setting_changed(
        self,
        _proxy: Gio.DBusProxy,
        _sender_name: str | None,
        signal_name: str,
        parameters: GLib.Variant,
        *_user_data: Any,
    ) -> None:
        if signal_name != "SettingChanged":
            return

        namespace, name, value = parameters
        if namespace == "org.freedesktop.appearance" and name == "color-scheme":
            self._prefer_dark = value == 1
            self._callback()
            app.ged.raise_event(StyleChanged())
            # Reload CSS to ensure mode-specific user colors are reloaded
            app.css_config.reload_css()

    @property
    def prefer_dark(self) -> bool | None:
        return self._prefer_dark



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing

import random

from nbxmpp.const import Affiliation
from nbxmpp.const import Role
from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common import types
from gajim.common.const import CONSONANTS
from gajim.common.const import VOWELS

if typing.TYPE_CHECKING:
    from gajim.common.modules.contacts import GroupchatOfflineParticipant
    from gajim.common.modules.contacts import GroupchatParticipant


def is_affiliation_change_allowed(
    self_contact: GroupchatParticipant,
    contact: GroupchatParticipant | GroupchatOfflineParticipant,
    target_aff: str | Affiliation,
) -> bool:
    if isinstance(target_aff, str):
        target_aff = Affiliation(target_aff)

    if contact.affiliation == target_aff:
        # Contact has already the target affiliation
        return False

    if self_contact.affiliation.is_owner:
        return True

    if not self_contact.affiliation.is_admin:
        return False

    if target_aff in (Affiliation.OWNER, Affiliation.ADMIN):
        # Admin can’t edit admin/owner list
        return False

    return self_contact.affiliation > contact.affiliation


def is_role_change_allowed(
    self_contact: GroupchatParticipant, contact: GroupchatParticipant
) -> bool:
    if self_contact.role < Role.MODERATOR:
        return False
    return self_contact.affiliation >= contact.affiliation


def is_moderation_allowed(self_contact: GroupchatParticipant) -> bool:
    return self_contact.role >= Role.MODERATOR


def get_default_muc_config() -> dict[str, bool | str]:
    return {
        # XEP-0045 options
        # https://xmpp.org/registrar/formtypes.html
        "muc#roomconfig_allowinvites": True,
        "muc#roomconfig_allowpm": "anyone",
        "muc#roomconfig_changesubject": False,
        "muc#roomconfig_enablelogging": False,
        "muc#roomconfig_membersonly": True,
        "muc#roomconfig_moderatedroom": False,
        "muc#roomconfig_passwordprotectedroom": False,
        "muc#roomconfig_persistentroom": True,
        "muc#roomconfig_publicroom": False,
        "muc#roomconfig_whois": "moderators",
        # Ejabberd options
        "allow_voice_requests": False,
        "public_list": False,
        "mam": True,
        # Prosody options
        "{http://prosody.im/protocol/muc}roomconfig_allowmemberinvites": False,
        "muc#roomconfig_enablearchiving": True,
    }


def get_random_muc_localpart() -> str:
    rand = random.randrange(4)
    is_vowel = bool(random.getrandbits(1))
    result = ""
    for _n in range(rand * 2 + (5 - rand)):
        if is_vowel:
            result = f"{result}{VOWELS[random.randrange(len(VOWELS))]}"
        else:
            result = f"{result}{CONSONANTS[random.randrange(len(CONSONANTS))]}"
        is_vowel = not is_vowel
    return result


def message_needs_highlight(text: str, nickname: str, own_jid: str) -> bool:
    """
    Check whether 'text' contains 'nickname', 'own_jid', or any string of the
    'muc_highlight_words' setting.
    """

    search_strings = app.settings.get("muc_highlight_words").split(";")
    search_strings.append(nickname)
    search_strings.append(own_jid)

    search_strings = [word.lower() for word in search_strings if word]
    text = text.lower()

    for search_string in search_strings:
        match = text.find(search_string)

        while match > -1:
            search_end = match + len(search_string)

            if match == 0 and search_end == len(text):
                # Text contains search_string only (exact match)
                return True

            # Exclude some characters preceding the match:
            # - any alpha chars
            # - / which may be commands
            # - - which may connect multiple words
            # - ' which may be part of a contraction, such as o'clock, 'tis
            excluded_chars = ("/", "-", "'")

            char_before = text[match - 1]
            char_before_allowed = bool(
                match == 0
                or (not char_before.isalpha() and char_before not in excluded_chars)
            )

            if char_before_allowed and search_end == len(text):
                # search_string found at the end of text and
                # char before search_string is allowed.
                return True

            if char_before_allowed and not text[search_end].isalpha():
                # char_before search_string is allowed and
                # char_after search_string is not alpha.
                return True

            start = match + 1
            match = text.find(search_string, start)

    return False


def get_groupchat_name(client: types.Client, jid: JID) -> str:
    name = client.get_module("Bookmarks").get_name_from_bookmark(jid)
    if name:
        return name

    disco_info = app.storage.cache.get_last_disco_info(jid)
    if disco_info is not None:
        if disco_info.muc_name:
            return disco_info.muc_name

    assert jid.localpart is not None
    return jid.localpart


def get_group_chat_nick(account: str, room_jid: JID | str) -> str:
    client = app.get_client(account)

    bookmark = client.get_module("Bookmarks").get_bookmark(room_jid)
    if bookmark is not None:
        if bookmark.nick is not None:
            return bookmark.nick

    return app.nicks[account]



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import Optional

import functools
import inspect
import sys
from collections.abc import Callable
from dataclasses import dataclass

from gi.repository import Gio
from gi.repository import GLib
from nbxmpp.protocol import JID

from gajim.common.structs import VariantMixin


@dataclass
class OpenEventActionParams(VariantMixin):
    # Event which is used for Notifications and gets sent over DBus
    # Don’t use Optional types here because DBus does not support "None"
    type: str
    sub_type: str
    account: str
    jid: str
    context_id: str


@dataclass
class AddChatActionParams(VariantMixin):
    account: str
    jid: JID
    type: Literal["chat", "groupchat", "pm"]
    select: bool


@dataclass
class ChatListEntryParam(VariantMixin):
    workspace_id: str
    source_workspace_id: str
    account: str
    jid: JID


@dataclass
class MuteContactParam(VariantMixin):
    account: str
    jid: JID
    state: int


@dataclass
class AccountJidParam(VariantMixin):
    account: str
    jid: JID


@dataclass
class OccupantParam(VariantMixin):
    account: str
    jid: JID
    resource: str
    occupant_id: str


@dataclass
class ExportHistoryParam(VariantMixin):
    account: str
    jid: Optional[JID]  # noqa: UP007, UP045


@dataclass
class ModerateMessageParam(VariantMixin):
    account: str
    jid: JID
    stanza_ids: list[str]
    namespace: str


@dataclass
class RetractMessageParam(VariantMixin):
    account: str
    jid: JID
    retract_ids: list[str]


@dataclass
class ModerateAllMessagesParam(VariantMixin):
    account: str
    jid: JID
    occupant_id: str
    nickname: str
    namespace: str


@dataclass
class DeleteMessageParam(VariantMixin):
    account: str
    jid: JID
    pk: int


@dataclass
class SubscriptionAcceptParam(VariantMixin):
    account: str
    jid: JID
    nickname: Optional[str]  # noqa: UP007, UP045


def get_params_class(func: Callable[..., Any]) -> Any:
    module = sys.modules[__name__]
    params = inspect.signature(func).parameters
    cls_string = params["params"].annotation
    cls_string = cls_string.rsplit(".", maxsplit=1)[-1]
    return getattr(module, cls_string)


def actionmethod(
    func: Callable[[Any, Gio.SimpleAction, Any], None],
) -> Callable[[Any, Gio.SimpleAction, GLib.Variant], None]:
    @functools.wraps(func)
    def method_wrapper(obj: Any, action: Gio.SimpleAction, param: GLib.Variant) -> None:
        params_cls = get_params_class(func)
        params = params_cls.from_variant(param)
        return func(obj, action, params)

    return method_wrapper


def actionfunction(
    func: Callable[[Gio.SimpleAction, Any], None],
) -> Callable[[Gio.SimpleAction, GLib.Variant], None]:
    @functools.wraps(func)
    def func_wrapper(action: Gio.SimpleAction, param: GLib.Variant) -> None:
        params_cls = get_params_class(func)
        params = params_cls.from_variant(param)
        return func(action, params)

    return func_wrapper



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from gi.repository import Gdk
from gi.repository import Gtk

from gajim.common import app
from gajim.common.client import BareContact
from gajim.common.const import AvatarSize
from gajim.common.i18n import _
from gajim.common.modules.chat_markers import DisplayedMarkerData
from gajim.common.modules.contacts import ResourceContact
from gajim.common.util.user_strings import get_uf_relative_time

from gajim.gtk.util.misc import container_remove_all
from gajim.gtk.util.misc import get_ui_string

MAX_AVATARS = 5


@Gtk.Template.from_string(string=get_ui_string("conversation/avatar_stack.ui"))
class AvatarStack(Gtk.MenuButton):
    __gtype_name__ = "AvatarStack"

    _avatar_box: Gtk.Box = Gtk.Template.Child()
    _more_label: Gtk.Label = Gtk.Template.Child()

    def __init__(self, account: str) -> None:
        Gtk.MenuButton.__init__(self)

        self._account = account
        self._client = app.get_client(self._account)

        self.set_cursor(Gdk.Cursor.new_from_name("pointer"))
        self.set_create_popup_func(self._on_clicked)

        self._scale_factor = self.get_scale_factor()

        self._markers: list[DisplayedMarkerData] = []

    def do_unroot(self) -> None:
        self.set_create_popup_func(None)
        Gtk.MenuButton.do_unroot(self)

    def set_data(self, markers: list[DisplayedMarkerData]) -> None:
        self._markers = markers.copy()
        container_remove_all(self._avatar_box)

        if not markers:
            return

        for entry in self._markers[:MAX_AVATARS]:
            self._avatar_box.append(self._get_avatar_image(entry))

        entries_count = len(self._markers)
        if entries_count > MAX_AVATARS:
            self._more_label.set_visible(True)
            self._more_label.set_label(f"+{entries_count - MAX_AVATARS}")
        else:
            self._more_label.set_visible(False)
            self._more_label.set_label("")

        marker_count = len(self._markers)
        if marker_count == 1:
            marker = markers[0]

            if marker.occupant is None:
                contact = self._client.get_module("Contacts").get_contact_if_exists(
                    marker.jid
                )
                assert contact is not None
                assert not isinstance(contact, ResourceContact)
                nickname = contact.name
            else:
                nickname = marker.occupant.nickname

            self.set_tooltip_text(_("Seen by %s") % nickname)
        else:
            self.set_tooltip_text(_("Seen by %s participants") % marker_count)

    def _get_avatar_image(self, marker: DisplayedMarkerData) -> Gtk.Image:
        texture = None
        if marker.occupant is None:
            contact = self._client.get_module("Contacts").get_contact_if_exists(
                marker.jid
            )
            assert contact is not None
            if isinstance(contact, BareContact):
                texture = contact.get_avatar(
                    size=AvatarSize.SMALL, scale=self._scale_factor, add_show=False
                )

        else:
            texture = app.app.avatar_storage.get_occupant_texture(
                marker.jid,
                marker.occupant,
                size=AvatarSize.SMALL,
                scale=self._scale_factor,
            )

        image = Gtk.Image.new_from_paintable(texture)
        image.set_pixel_size(AvatarSize.SMALL)
        return image

    def _on_clicked(self, _widget: AvatarStack) -> None:
        self.set_popover(AvatarStackPopover(self._account, self._markers))


@Gtk.Template.from_string(string=get_ui_string("conversation/avatar_stack_popover.ui"))
class AvatarStackPopover(Gtk.Popover):
    __gtype_name__ = "AvatarStackPopover"

    _listbox: Gtk.ListBox = Gtk.Template.Child()

    def __init__(self, account: str, data: list[DisplayedMarkerData]) -> None:
        Gtk.Popover.__init__(self)

        for entry in data:
            self._listbox.append(AvatarStackPopoverRow(account, entry))


@Gtk.Template.from_string(
    string=get_ui_string("conversation/avatar_stack_popover_row.ui")
)
class AvatarStackPopoverRow(Gtk.ListBoxRow):
    __gtype_name__ = "AvatarStackPopoverRow"

    _avatar_image: Gtk.Image = Gtk.Template.Child()
    _contact_name_label: Gtk.Label = Gtk.Template.Child()
    _timestamp_label: Gtk.Label = Gtk.Template.Child()

    def __init__(self, account: str, marker: DisplayedMarkerData) -> None:
        Gtk.ListBoxRow.__init__(self)

        nickname = ""
        texture = None

        if marker.occupant is None:
            client = app.get_client(account)
            contact = client.get_module("Contacts").get_contact_if_exists(marker.jid)
            assert contact is not None
            if isinstance(contact, BareContact):
                texture = contact.get_avatar(
                    size=AvatarSize.SMALL, scale=self.get_scale_factor(), add_show=False
                )
                nickname = contact.name

        else:
            texture = app.app.avatar_storage.get_occupant_texture(
                marker.jid,
                marker.occupant,
                size=AvatarSize.ROSTER,
                scale=self.get_scale_factor(),
            )
            nickname = marker.occupant.nickname or ""

        self._avatar_image.set_from_paintable(texture)
        self._avatar_image.set_pixel_size(AvatarSize.ROSTER)

        self._contact_name_label.set_text(nickname)

        self._timestamp_label.set_text(get_uf_relative_time(marker.timestamp))



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import datetime as dt
import logging

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango
from nbxmpp.structs import Hat

from gajim.common import app
from gajim.common.i18n import _
from gajim.common.util.user_strings import format_idle_time
from gajim.common.util.user_strings import get_uf_relative_time

from gajim.gtk.util.misc import container_remove_all

# Backwards compat import for Plugins
from gajim.gtk.window import GajimAppWindow as GajimAppWindow  # noqa: PLC0414

log = logging.getLogger("gajim.gtk.widgets")


class MultiLineLabel(Gtk.Label):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        Gtk.Label.__init__(self, *args, **kwargs)
        self.set_wrap(True)
        self.set_wrap_mode(Pango.WrapMode.WORD_CHAR)
        self.set_single_line_mode(False)
        self.set_selectable(True)


class GroupBadge(Gtk.Label):
    def __init__(self, group: str) -> None:
        Gtk.Label.__init__(
            self,
            ellipsize=Pango.EllipsizeMode.END,
            halign=Gtk.Align.END,
            hexpand=True,
            label=group,
            valign=Gtk.Align.CENTER,
            max_width_chars=20,
        )

        self.add_css_class("badge")
        self.add_css_class("badge-group")


class GroupBadgeBox(Gtk.Box):
    __gtype_name__ = "GroupBadgeBox"

    def __init__(self) -> None:
        Gtk.Box.__init__(self)

        self._groups: list[str] = []

    @GObject.Property(type=object)
    def groups(self) -> list[str]:  # pyright: ignore
        return self._groups

    @groups.setter
    def groups(self, groups: list[str] | None) -> None:
        if groups is None:
            groups = []

        self._groups = groups[:3]
        container_remove_all(self)
        visible = bool(groups)
        self.set_visible(visible)
        if not visible:
            return

        for group in self._groups:
            self.append(GroupBadge(group))


class HatBadge(Gtk.Box):
    def __init__(self, hat: Hat) -> None:
        Gtk.Box.__init__(self, spacing=6, halign=Gtk.Align.START)
        self.add_css_class("badge")
        self.add_css_class("badge-hat")
        if hat.hue is not None:
            css_provider = Gtk.CssProvider()
            css_provider.load_from_string(
                f".badge-hat {{background-color: hsl({hat.hue}, 100%, 25%);}}"
            )
            context = self.get_style_context()
            context.add_provider(css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        image = Gtk.Image.new_from_icon_name("lucide-tag-symbolic")
        image.set_pixel_size(12)
        self.append(image)

        label = Gtk.Label(
            label=GLib.markup_escape_text(hat.title),
            ellipsize=Pango.EllipsizeMode.END,
            max_width_chars=20,
            halign=Gtk.Align.START,
        )
        self.append(label)


class IdleBadge(Gtk.Label):
    __gtype_name__ = "IdleBadge"

    def __init__(self, idle: dt.datetime | None = None) -> None:
        Gtk.Label.__init__(
            self,
            halign=Gtk.Align.START,
            hexpand=True,
            ellipsize=Pango.EllipsizeMode.NONE,
            visible=False,
        )
        self.set_size_request(50, -1)
        self.add_css_class("dimmed")
        self.add_css_class("caption")

    @GObject.Property(type=object)
    def idle(self) -> str:  # pyright: ignore
        return self.get_text()

    @idle.setter
    def idle(self, value: dt.datetime | None) -> None:
        return self._set_idle(value)

    def _set_idle(self, value: dt.datetime | None) -> None:
        self.set_visible(bool(value))
        if value is None:
            return

        self.set_text(_("Last seen: %s") % format_idle_time(value))
        format_string = app.settings.get("date_time_format")
        self.set_tooltip_text(value.strftime(format_string))


class AccountBadge(Gtk.Label):
    __gtype_name__ = "AccountBadge"

    def __init__(self, account: str | None = None, bind_setting: bool = False) -> None:
        Gtk.Label.__init__(self)
        self.set_ellipsize(Pango.EllipsizeMode.END)
        self.set_max_width_chars(12)
        self.set_size_request(50, -1)
        self.add_css_class("badge")

        self._bind_setting = bind_setting

        if account is not None:
            self.set_account(account)

        self.set_visible(account is not None)

    def do_unroot(self) -> None:
        app.settings.disconnect_signals(self)
        Gtk.Label.do_unroot(self)
        app.check_finalize(self)

    @GObject.Property(type=str)
    def account(self) -> str:  # pyright: ignore
        return self.get_text()

    @account.setter
    def account(self, value: str) -> None:
        self.set_account(value)

    def set_account(self, account: str) -> None:
        label = app.get_account_label(account)
        self.set_text(label)

        for style_class in self.get_css_classes():
            if style_class.startswith("gajim_class"):
                self.remove_css_class(style_class)

        account_class = app.css_config.get_dynamic_class(account)
        self.add_css_class(account_class)

        self.set_tooltip_text(_("Account: %s") % label)
        if self._bind_setting:
            app.settings.disconnect_signals(self)
            app.settings.connect_signal(
                "account_label", self._on_account_label_changed, account
            )

    def _on_account_label_changed(
        self, _value: str, _setting: str, account: str | None, *args: Any
    ) -> None:
        assert account is not None
        self.set_account(account)


class GdkRectangle(Gdk.Rectangle):
    def __init__(self, x: int, y: int, height: int = 1, width: int = 1) -> None:
        Gdk.Rectangle.__init__(self)
        self.x = x
        self.y = y
        self.height = height
        self.width = width


class GajimPopover(Gtk.PopoverMenu):
    __gtype_name__ = "GajimPopover"

    def __init__(
        self,
        menu: Gio.MenuModel | None = None,
        position: Gtk.PositionType = Gtk.PositionType.RIGHT,
        event: Any | None = None,
    ) -> None:
        Gtk.PopoverMenu.__init__(self, autohide=True)

        if menu is not None:
            self.set_menu_model(menu)

        self.set_position(position)
        if event is not None:
            self.set_pointing_from_event(event)

    def set_pointing_from_event(self, event: Any) -> None:
        self.set_pointing_to_coord(event.x, event.y)

    def set_pointing_to_coord(self, x: float, y: float) -> None:
        rectangle = GdkRectangle(x=int(x), y=int(y))
        self.set_pointing_to(rectangle)


class TimeLabel(Gtk.Label):
    __gtype_name__ = "TimeLabel"

    def __init__(self) -> None:
        Gtk.Label.__init__(self)
        self._timestamp = None
        app.pulse_manager.add_callback(self.pulse)

    def do_unroot(self) -> None:
        app.pulse_manager.remove_callback(self.pulse)
        Gtk.Label.do_unroot(self)

    def set_timestamp(self, timestamp: dt.datetime) -> None:
        self._timestamp = timestamp
        self.pulse()

    def pulse(self) -> None:
        if self._timestamp is None:
            return
        rel_string = get_uf_relative_time(self._timestamp)
        self.set_text(rel_string)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gajim.common import types
from gajim.common.modules.base import BaseModule


class AdHocCommands(BaseModule):

    _nbxmpp_extends = 'AdHoc'
    _nbxmpp_methods = [
        'request_command_list',
        'execute_command',
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# Based on https://learn.microsoft.com/en-us/windows/apps/desktop/modernize/apply-windows-themes

from __future__ import annotations

from typing import Any

import logging
from collections.abc import Callable

from winrt.windows.ui import Color
from winrt.windows.ui.viewmanagement import UIColorType
from winrt.windows.ui.viewmanagement import UISettings

from gajim.common import app
from gajim.common.events import StyleChanged

log = logging.getLogger("gajim.c.winapi.system_style")


class SystemStyleListener:
    def __init__(self, callback: Callable[..., None]) -> None:
        self._prefer_dark: bool | None = None
        self._callback = callback

        try:
            self._ui_settings = UISettings()
            self._ui_settings.add_color_values_changed(
                self._signal_color_values_changed
            )
            self._prefer_dark = self._get_prefer_dark()
        except Exception as error:
            log.warning("Failed to init winsdk.UISettings: %s", error)
            return

    def _get_prefer_dark(self) -> bool:
        foreground_color = self._ui_settings.get_color_value(UIColorType.FOREGROUND)
        return self._is_color_light(foreground_color)

    @staticmethod
    def _is_color_light(clr: Color) -> bool:
        return ((5 * clr.g) + (2 * clr.r) + clr.b) > (8 * 128)

    def _signal_color_values_changed(
        self, _ui_settings: UISettings | None, *args: Any
    ) -> None:
        prefer_dark = self._get_prefer_dark()
        if prefer_dark != self._prefer_dark:
            self._prefer_dark = prefer_dark
            self._callback()
            app.ged.raise_event(StyleChanged())
            # Reload CSS to ensure mode-specific user colors are reloaded
            app.css_config.reload_css()

    @property
    def prefer_dark(self) -> bool | None:
        return self._prefer_dark



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common import types
from gajim.common.const import SimpleClientState
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant

from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string


@Gtk.Template.from_string(string=get_ui_string("contact_name_entry.ui"))
class ContactNameEntry(Gtk.Box, SignalManager):
    __gtype_name__ = "ContactNameEntry"

    __gsignals__ = {
        "name-updated": (GObject.SignalFlags.RUN_LAST, None, (str,)),
    }

    _name_label: Gtk.Label = Gtk.Template.Child()
    _edit_button: Gtk.Button = Gtk.Template.Child()

    _name_entry_box: Gtk.Box = Gtk.Template.Child()
    _name_entry: Gtk.Entry = Gtk.Template.Child()
    _apply_button: Gtk.Button = Gtk.Template.Child()
    _cancel_button: Gtk.Button = Gtk.Template.Child()
    _reset_button: Gtk.Button = Gtk.Template.Child()

    def __init__(
        self,
        contact: BareContact | GroupchatContact | GroupchatParticipant | None = None,
        editable: bool = False,
    ) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        self._contact = contact

        name = "" if self._contact is None else self._contact.name

        self.update_displayed_name(name)

        self._edit_button.set_visible(editable)
        self._connect(self._edit_button, "clicked", self._on_edit_clicked)
        if (
            isinstance(self._contact, GroupchatParticipant)
            or (
                isinstance(self._contact, BareContact)
                and not self._contact.is_in_roster
            )
            or not editable
        ):
            self._edit_button.set_visible(False)

        self._connect(self._name_entry, "activate", self._on_entry_activated)
        self._connect(self._apply_button, "clicked", self._on_apply_clicked)
        self._connect(self._cancel_button, "clicked", self._on_cancel_clicked)
        self._connect(self._reset_button, "clicked", self._on_reset_clicked)

        if self._contact is not None:
            client = app.get_client(self._contact.account)
            client.connect_signal("state-changed", self._on_client_state_changed)

    def do_unroot(self) -> None:
        Gtk.Box.do_unroot(self)
        self._disconnect_all()
        app.check_finalize(self)

    def _on_client_state_changed(
        self, _client: types.Client, _signal_name: str, state: SimpleClientState
    ) -> None:
        self._edit_button.set_sensitive(state.is_connected)
        self._apply_button.set_sensitive(state.is_connected)
        self._reset_button.set_sensitive(state.is_connected)

    def _on_edit_clicked(self, _button: Gtk.Button) -> None:
        self.enable_edit_mode()

    def _on_entry_activated(self, _entry: Gtk.Entry) -> None:
        self._save_name()
        self._disable_edit_mode()

    def _on_apply_clicked(self, _button: Gtk.Button) -> None:
        self._save_name()
        self._disable_edit_mode()

    def _on_cancel_clicked(self, _button: Gtk.Button) -> None:
        self._disable_edit_mode()

    def _on_reset_clicked(self, _button: Gtk.Button) -> None:
        self._name_entry.set_text("")
        self._save_name()
        self._disable_edit_mode()

    def _save_name(self) -> None:
        assert isinstance(self._contact, BareContact | GroupchatContact)

        name = self._name_entry.get_text()
        client = app.get_client(self._contact.account)

        if isinstance(self._contact, BareContact):
            client.get_module("Roster").change_name(self._contact.jid, name)
        else:
            client.get_module("Bookmarks").modify(self._contact.jid, name=name)

        self.update_displayed_name(name)

    def _disable_edit_mode(self) -> None:
        self._name_entry_box.set_visible(False)
        self._cancel_button.set_visible(False)
        self._reset_button.set_visible(False)

        self._name_label.set_visible(True)
        self._edit_button.set_visible(True)

    def enable_edit_mode(self) -> None:
        if self._contact is None:
            return

        client = app.get_client(self._contact.account)
        if not client.state.is_connected and not client.state.is_available:
            return

        self._name_label.set_visible(False)
        self._edit_button.set_visible(False)

        self._name_entry_box.set_visible(True)

        self._name_entry.set_text(self._contact.name)
        self._name_entry.grab_focus()

        self._cancel_button.set_visible(True)
        self._reset_button.set_visible(True)

    def update_displayed_name(self, name: str) -> None:
        if self._contact is not None:
            if isinstance(self._contact, BareContact | GroupchatContact):
                # Name editing is not possible for GroupchatParticipant
                if name == "":
                    name = self._contact.original_name

                if name != self._contact.original_name:
                    name = f"{name} ({self._contact.original_name})"
            else:
                name = self._contact.name

        self._name_label.set_label(name)
        self._name_label.set_tooltip_text(name)

        self._name_entry.set_text(name)

        self.emit("name-updated", name)

    def set_contact(
        self, contact: BareContact | GroupchatContact | GroupchatParticipant
    ) -> None:
        self._contact = contact
        self.update_displayed_name(self._contact.name)



# Copyright (C) 2006 Jean-Marie Traissard <jim AT lapin.org>
#                    Junglecow J <junglecow AT gmail.com>
# Copyright (C) 2006-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2007 Brendan Taylor <whateley AT gmail.com>
# Copyright (C) 2008 Jonathan Schleifer <js-gajim AT webkeks.org>
# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import cast

import importlib.resources
import os
import shutil
import sys
import tempfile
from collections.abc import Generator
from pathlib import Path

from gi.repository import GLib

import gajim
from gajim.common.const import PathLocation
from gajim.common.const import PathType
from gajim.common.i18n import _

PathTupleT = tuple[PathLocation | None, Path, PathType | None]


def get(key: str) -> Path:
    return _paths[key]


def get_temp_dir() -> Path:
    return _paths.get_temp_dir()


def get_plugin_dirs() -> list[Path]:
    if gajim.IS_FLATPAK:
        return [Path(_paths['PLUGINS_BASE']),
                Path('/app/plugins')]
    return [Path(_paths['PLUGINS_BASE']),
            Path(_paths['PLUGINS_USER'])]


def get_ui_path(filename: str) -> Path:
    return _paths['GUI'] / filename


def get_paths(type_: PathType) -> Generator[Path, None, None]:
    for key, value in _paths.items():
        path_type = value[2]
        if type_ != path_type:
            continue
        yield _paths[key]  # noqa: PLR1733


def set_separation(active: bool) -> None:
    # Deprecated in Gajim 2.3.0
    _paths.profile_separation = active


def set_profile(profile: str) -> None:
    # Deprecated in Gajim 2.3.0
    _paths.profile = profile


def set_user_profile(user_profile: str) -> None:
    _paths.user_profile = user_profile


def set_config_root(config_root: str) -> None:
    _paths.custom_config_root = Path(config_root).resolve()


def init() -> None:
    _paths.init()


def create_paths() -> None:
    for path in get_paths(PathType.FOLDER):
        if path.is_file():
            print(_('%s is a file but it should be a directory') % path)
            print(_('Gajim will now exit'))
            sys.exit()

        if not path.exists():
            for parent_path in reversed(path.parents):
                # Create all parent folders
                # don't use mkdir(parent=True), as it ignores `mode`
                # when creating the parents
                if not parent_path.exists():
                    print(('creating %s directory') % parent_path)
                    parent_path.mkdir(mode=0o700)
            print(('creating %s directory') % path)
            path.mkdir(mode=0o700)


def cleanup_temp() -> None:
    tmpdir = _paths.get_temp_dir().parent
    for path in tmpdir.glob('gajim-*'):
        if not path.is_dir():
            continue
        try:
            shutil.rmtree(path, ignore_errors=True)
        except Exception:
            pass


class ConfigPaths:
    def __init__(self) -> None:
        self._paths: dict[str, PathTupleT] = {}
        self._temp_dir: Path | None = None

        self.profile = ''
        self.user_profile = ''
        self.profile_separation = False
        self.custom_config_root: Path | None = None


        basedir = cast(Path, importlib.resources.files('gajim'))

        source_paths = [
            ('DATA', basedir / 'data'),
            ('STYLE', basedir / 'data' / 'style'),
            ('GUI', basedir / 'data' / 'gui'),
            ('ICONS', basedir / 'data' / 'icons'),
            ('HOME', Path.home()),
            ('PLUGINS_BASE', basedir / 'data' / 'plugins'),
        ]

        for path in source_paths:
            self.add(*path)

    def __getitem__(self, key: str) -> Path:
        location, path, _ = self._paths[key]
        if location == PathLocation.CONFIG:
            return self.config_root / path
        if location == PathLocation.CACHE:
            return self.cache_root / path
        if location == PathLocation.DATA:
            return self.data_root / path
        return path

    def items(self) -> Generator[tuple[str, PathTupleT], None, None]:
        yield from self._paths.items()

    def _prepare(self, path: Path, unique: bool) -> Path:
        if os.name == 'nt':
            path = Path(str(path).capitalize())
        if self.profile:
            if unique or self.profile_separation:
                return Path(f'{path}.{self.profile}')
        return path

    def add(self,
            name: str,
            path: Path | str,
            location: PathLocation | None = None,
            path_type: PathType | None = None,
            unique: bool = False) -> None:

        path = Path(path)

        if location != PathLocation.NONE:
            path = self._prepare(path, unique)
        self._paths[name] = (location, path, path_type)

    def init(self):

        root_folder = 'gajim'
        if self.user_profile:
            root_folder = f'gajim.{self.user_profile}'

        if sys.platform == 'win32':
            if gajim.IS_PORTABLE:

                root_folder = 'UserData'
                if self.user_profile:
                    root_folder = f'UserData.{self.user_profile}'

                application_path = Path(sys.executable).parent
                self.config_root = self.cache_root = self.data_root = \
                    application_path.parent / root_folder
            else:
                # Documents and Settings\[User Name]\Application Data\Gajim
                self.config_root = self.cache_root = self.data_root = \
                    Path(os.environ['APPDATA']) / root_folder.capitalize()
        else:
            self.config_root = Path(GLib.get_user_config_dir()) / root_folder
            self.cache_root = Path(GLib.get_user_cache_dir()) / root_folder
            self.data_root = Path(GLib.get_user_data_dir()) / root_folder

        if self.custom_config_root:
            self.config_root = self.custom_config_root
            self.cache_root = self.custom_config_root
            self.data_root = self.custom_config_root

        user_dir_paths = [
            ('MY_CONFIG', Path(), PathLocation.CONFIG, PathType.FOLDER),
            ('MY_CACHE', Path(), PathLocation.CACHE, PathType.FOLDER),
            ('MY_DATA', Path(), PathLocation.DATA, PathType.FOLDER),
        ]

        for path in user_dir_paths:
            self.add(*path)

        # These paths are unique per profile
        unique_profile_paths: list[tuple[str, str | Path, PathLocation, PathType]] = [
            # Data paths
            ('CERT_STORE', 'cert_store', PathLocation.DATA, PathType.FOLDER),
            ('DEBUG', 'debug', PathLocation.DATA, PathType.FOLDER),
            ('PLUGINS_DATA', 'plugins_data',
             PathLocation.DATA, PathType.FOLDER),

            # Cache paths
            ('DOWNLOADS_THUMB', 'downloads.thumb', PathLocation.CACHE, PathType.FOLDER),

            # Config paths
            ('SETTINGS', 'settings.sqlite', PathLocation.CONFIG, PathType.FILE),
            ('PLUGINS_CONFIG_DIR',
             'pluginsconfig', PathLocation.CONFIG, PathType.FOLDER),
            ('MY_SHORTCUTS', 'shortcuts.json',
             PathLocation.CONFIG, PathType.FILE),
        ]

        # Determine downloads dir
        path = ('DOWNLOADS', 'downloads', PathLocation.DATA, PathType.FOLDER)
        if sys.platform == 'win32' and not gajim.IS_PORTABLE:
            download_dir = GLib.get_user_special_dir(
                GLib.UserDirectory.DIRECTORY_DOWNLOAD)
            assert download_dir is not None
            path = ('DOWNLOADS', Path(download_dir) / 'Gajim',
                    PathLocation.NONE, PathType.FOLDER)

        unique_profile_paths.append(path)

        for path in unique_profile_paths:
            self.add(*path, unique=True)

        # These paths are only unique per profile if the commandline arg
        # `separate` is passed
        paths = [
            # Data paths
            ('LOG_DB', 'logs.db', PathLocation.DATA, PathType.FILE),
            ('PLUGINS_DOWNLOAD', 'plugins_download',
             PathLocation.CACHE, PathType.FOLDER),
            ('PLUGINS_IMAGES', 'plugins_images',
             PathLocation.CACHE, PathType.FOLDER),
            ('PLUGINS_USER', 'plugins', PathLocation.DATA, PathType.FOLDER),
            ('MY_ICONSETS',
             'iconsets', PathLocation.DATA, PathType.FOLDER_OPTIONAL),

            # Cache paths
            ('CACHE_DB', 'cache.db', PathLocation.CACHE, PathType.FILE),
            ('AVATAR', 'avatars', PathLocation.CACHE, PathType.FOLDER),
            ('AVATAR_ICONS', 'avatar_icons',
             PathLocation.CACHE, PathType.FOLDER),
            ('BOB', 'bob', PathLocation.CACHE, PathType.FOLDER),

            # Config paths
            ('MY_THEME', 'theme', PathLocation.CONFIG, PathType.FOLDER),

        ]

        for path in paths:
            self.add(*path)

    def get_temp_dir(self) -> Path:
        if self._temp_dir is None or not self._temp_dir.exists():
            self._temp_dir = Path(tempfile.mkdtemp(prefix='gajim-'))
        return self._temp_dir


_paths = ConfigPaths()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import datetime as dt
import itertools
import logging
from collections.abc import Iterator
from dataclasses import dataclass
from enum import Enum

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp import JID

from gajim.common import app
from gajim.common import ged
from gajim.common.client import Client
from gajim.common.const import AvatarSize
from gajim.common.const import Direction
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.modules.contacts import ResourceContact
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageType
from gajim.common.storage.archive.models import Message

from gajim.gtk.builder import get_builder
from gajim.gtk.conversation.message_widget import MessageWidget
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import convert_py_to_glib_datetime

log = logging.getLogger("gajim.gtk.search_view")


class PlaceholderMode(Enum):
    INITIAL = "initial"
    SEARCHING = "searching"
    NO_RESULTS = "no_results"


class SearchView(Gtk.Box, SignalManager, EventHelper):
    __gtype_name__ = "SearchView"

    __gsignals__ = {
        "hide-search": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self) -> None:
        Gtk.Box.__init__(self, width_request=300)
        SignalManager.__init__(self)
        EventHelper.__init__(self)

        self._account: str | None = None
        self._jid: JID | None = None
        self._results_iterator: Iterator[Message] | None = None

        self._first_date: dt.datetime | None = None
        self._last_date: dt.datetime | None = None

        self._last_search_string = ""

        self._ui = get_builder("search_view.ui", ["search_box"])
        self._ui.results_listbox.set_header_func(self._header_func)
        self.append(self._ui.search_box)

        self._search_filters = SearchFilters()
        self._connect(
            self._search_filters, "filter-changed", self._on_search_filters_changed
        )
        self._connect(
            self._search_filters, "filter-activated", self._on_search_filters_activated
        )
        self._ui.search_filters_box.append(self._search_filters)

        self.register_events(
            [
                ("account-enabled", ged.GUI1, self._on_account_state),
                ("account-disabled", ged.GUI1, self._on_account_state),
            ]
        )

        self._connect(self._ui.calendar, "day-selected", self._on_date_selected)
        self._connect(self._ui.calendar, "notify::month", self._on_month_changed)
        self._connect(self._ui.calendar, "notify::year", self._on_month_changed)

        self._connect(
            self._ui.calendar_button, "notify::active", self._on_calender_button_clicked
        )
        self._connect(
            self._ui.first_day_button, "clicked", self._on_first_date_selected
        )
        self._connect(
            self._ui.previous_day_button, "clicked", self._on_previous_date_selected
        )
        self._connect(self._ui.next_day_button, "clicked", self._on_next_date_selected)
        self._connect(self._ui.last_day_button, "clicked", self._on_last_date_selected)
        self._connect(self._ui.close_button, "clicked", self._on_hide_clicked)
        self._connect(self._ui.search_entry, "activate", self._on_search)
        self._connect(
            self._ui.search_checkbutton, "toggled", self._on_search_all_toggled
        )
        self._connect(self._ui.results_scrolled, "edge-reached", self._on_edge_reached)
        self._connect(self._ui.results_listbox, "row-activated", self._on_row_activated)

    def do_unroot(self) -> None:
        self._disconnect_all()
        self.unregister_events()
        app.check_finalize(self)

    def _on_account_state(self, _event: Any) -> None:
        self._clear()

    @staticmethod
    def _header_func(row: ResultRow, before: ResultRow | None) -> None:
        if before is None:
            row.set_header(RowHeader(row.account, row.remote_jid, row.timestamp))
        else:
            if before.remote_jid != row.remote_jid:
                row.set_header(RowHeader(row.account, row.remote_jid, row.timestamp))
            elif before.local_timestamp.date() != row.local_timestamp.date():
                row.set_header(RowHeader(row.account, row.remote_jid, row.timestamp))
            else:
                row.set_header(None)

    def _on_hide_clicked(self, _button: Gtk.Button) -> None:
        self.emit("hide-search")
        self._clear()

    def _clear(self) -> None:
        self._last_search_string = ""
        self._ui.search_entry.set_text("")
        self._search_filters.reset()
        self._clear_results()

    def _clear_results(self) -> None:
        # Unset the header_func to reduce load when clearing
        self._ui.results_listbox.set_header_func(None)

        self._ui.results_listbox.remove_all()
        # Set placeholder again, otherwise it won't be shown
        self._ui.results_listbox.set_placeholder(self._ui.status_page)

        self._ui.results_listbox.set_header_func(self._header_func)
        self._ui.results_scrolled.get_vadjustment().set_value(0)

        self._set_placeholder_mode(PlaceholderMode.INITIAL)

    def _on_search_all_toggled(self, _checkbutton: Gtk.CheckButton) -> None:
        # Reset state to allow changing scope while not changing search string
        self._last_search_string = ""

    def _on_search_filters_changed(self, _search_filters: SearchFilters) -> None:
        # Reset search string to allow new searches after changed filters
        self._last_search_string = ""

    def _on_search_filters_activated(self, _search_filters: SearchFilters) -> None:
        self._ui.search_entry.activate()

    def _on_search(self, entry: Gtk.Entry) -> None:
        text = entry.get_text()
        if text == self._last_search_string:
            # Return early if search string did not change
            # (prevents burst of db queries when holding enter).
            return

        self._last_search_string = text

        self._clear_results()
        if not text:
            return

        everywhere = self._ui.search_checkbutton.get_active()
        context = self._account is not None and self._jid is not None
        if not context:
            # Started search without context -> show in UI
            self._ui.search_checkbutton.set_active(True)

        if not context or everywhere:
            account = None
            jid = None
        else:
            account = self._account
            jid = self._jid

        search_filter = self._search_filters.get_filters()

        self._set_placeholder_mode(PlaceholderMode.SEARCHING)
        self._results_iterator = app.storage.archive.search_archive(
            account,
            jid,
            text,
            from_users=search_filter.usernames,
            before=search_filter.before,
            after=search_filter.after,
        )

        self._add_results()

    def _set_placeholder_mode(self, placeholder_mode: PlaceholderMode) -> None:
        if placeholder_mode == PlaceholderMode.SEARCHING:
            self._ui.status_page.set_title(_("Searching…"))
            self._ui.status_page.set_description(None)
            self._ui.status_page.set_paintable(
                Adw.SpinnerPaintable(widget=self._ui.status_page)
            )
            return

        if placeholder_mode == PlaceholderMode.NO_RESULTS:
            text = _("No Results Found")
            description = _("Try a different search")
        else:
            # PlaceholderMode.INITIAL
            text = _("Search")
            description = _("Search in chats…")

        self._ui.status_page.set_icon_name("lucide-search-symbolic")
        self._ui.status_page.set_title(text)
        self._ui.status_page.set_description(description)

    def _add_results(self) -> None:
        assert self._results_iterator is not None
        has_results = False
        for message in itertools.islice(self._results_iterator, 25):
            # With the current database design for corrections, we found
            # no way of only searching within the last correction of a
            # message so the search can return the original message, the
            # last correction or any in between correction.

            if message.get_last_correction() is not None:
                # This is only true for original messages which have
                # corrections, dont show them because they are obsolete
                continue

            if message.correction_id:
                # This is only true for the correction of a message.
                # We need to find out if this correction is obsolete by
                # checking if this correction is the last correction

                original_message = app.storage.archive.get_corrected_message(message)
                if original_message is None:
                    continue

                last_correction = original_message.get_last_correction()
                assert last_correction is not None
                if message.pk != last_correction.pk:
                    # This was a search hit on a correction which is not
                    # the last correction available, so ignore it
                    continue

            result_row = ResultRow(message)
            self._ui.results_listbox.append(result_row)
            has_results = True

        if not has_results:
            self._set_placeholder_mode(PlaceholderMode.NO_RESULTS)

    def _on_edge_reached(
        self, _scrolledwin: Gtk.ScrolledWindow, pos: Gtk.PositionType
    ) -> None:
        if pos != Gtk.PositionType.BOTTOM:
            return

        self._add_results()

    def _on_calender_button_clicked(
        self, menu_button: Gtk.MenuButton, *args: Any
    ) -> None:
        if menu_button.get_active():
            self._update_calendar()

    def _update_calendar(self) -> None:
        assert self._jid is not None
        assert self._account is not None

        first_log = app.storage.archive.get_first_message_ts(self._account, self._jid)
        if first_log is None:
            return
        self._first_date = first_log.astimezone()
        last_log = app.storage.archive.get_last_message_ts(self._account, self._jid)
        if last_log is None:
            return
        self._last_date = last_log.astimezone()

        self._ui.calendar.select_day(convert_py_to_glib_datetime(self._last_date))
        self._update_marks()

    def _on_month_changed(self, _calendar: Gtk.Calendar, *args: Any) -> None:
        self._update_marks()

    def _on_year_changed(self, _calendar: Gtk.Calendar, *args: Any) -> None:
        self._update_marks()

    def _update_marks(self) -> None:
        # Mark days with history in calendar
        date_time = self._ui.calendar.get_date()
        if date_time.get_year() < 1900:
            new_date_time = GLib.DateTime.new_from_iso8601("1900-01-01T00:00:00Z")
            assert new_date_time
            self._ui.calendar.select_day(new_date_time)
            return

        self._ui.calendar.clear_marks()

        assert self._jid is not None
        assert self._account is not None

        history_days = app.storage.archive.get_days_containing_messages(
            self._account, self._jid, date_time.get_year(), date_time.get_month()
        )
        for day in history_days:
            self._ui.calendar.mark_day(day)

    def _on_date_selected(self, calendar: Gtk.Calendar) -> None:
        date_time = calendar.get_date()
        date = dt.datetime(*date_time.get_ymd())
        app.window.get_control().scroll_to_date(date)

    def _on_first_date_selected(self, _button: Gtk.Button) -> None:
        assert self._first_date is not None
        self._ui.calendar.select_day(convert_py_to_glib_datetime(self._first_date))

    def _on_last_date_selected(self, _button: Gtk.Button) -> None:
        assert self._last_date is not None
        self._ui.calendar.select_day(convert_py_to_glib_datetime(self._last_date))

    def _on_previous_date_selected(self, _button: Gtk.Button) -> None:
        delta = dt.timedelta(days=-1)
        assert self._first_date is not None
        self._select_date(delta, self._first_date.date(), Direction.NEXT)

    def _on_next_date_selected(self, _button: Gtk.Button) -> None:
        delta = dt.timedelta(days=1)
        assert self._last_date is not None
        self._select_date(delta, self._last_date.date(), Direction.PREV)

    def _select_date(
        self, delta: dt.timedelta, end_date: dt.date, direction: Direction
    ) -> None:
        # Iterate through days until history entry found or
        # supplied end_date (first_date/last_date) reached

        g_datetime = self._ui.calendar.get_date()
        date = dt.date(*g_datetime.get_ymd())

        assert self._jid is not None
        assert self._account is not None

        day_has_messages = False
        while not day_has_messages:
            if direction == Direction.PREV:
                if end_date <= date:
                    return
            else:
                if end_date >= date:
                    return

            date = date + delta
            if date == end_date:
                break
            day_has_messages = self._ui.calendar.get_day_is_marked(date.day)

        self._ui.calendar.select_day(convert_py_to_glib_datetime(date))

    @staticmethod
    def _on_row_activated(_listbox: SearchView, row: ResultRow) -> None:
        app.window.scroll_to_message(row.account, row.message)

    def set_focus(self) -> None:
        self._ui.search_entry.grab_focus()
        self._clear()

    def set_context(self, account: str | None, jid: JID | None) -> None:
        self._account = account
        self._jid = jid
        self._last_search_string = ""

        self._search_filters.set_context(account, jid)

        if self._account is None and self._jid is None:
            self._ui.calendar_button.set_sensitive(False)
            return

        self._ui.calendar_button.set_sensitive(True)


class RowHeader(Gtk.Box):
    def __init__(self, account: str, jid: JID, timestamp: dt.datetime) -> None:
        Gtk.Box.__init__(self)
        self.set_hexpand(True)

        self._ui = get_builder("search_view.ui", ["header_box"])
        self.append(self._ui.header_box)

        client = app.get_client(account)
        contact = client.get_module("Contacts").get_contact(jid)
        assert isinstance(
            contact, BareContact | GroupchatContact | GroupchatParticipant
        )
        self._ui.header_name_label.set_text(contact.name or "")

        local_timestamp = timestamp.astimezone()

        format_string = app.settings.get("time_format")
        if local_timestamp.date() <= dt.datetime.today().date():
            format_string = app.settings.get("date_format")
        self._ui.header_date_label.set_text(local_timestamp.strftime(format_string))


class ResultRow(Gtk.ListBoxRow):
    def __init__(self, message: Message) -> None:
        Gtk.ListBoxRow.__init__(self)

        self._client = self._get_client(str(message.account.jid))
        self.account = self._client.account

        self.message = message
        self.remote_jid = message.remote.jid
        self.direction = ChatDirection(message.direction)

        jid = message.remote.jid
        if message.direction == ChatDirection.OUTGOING:
            jid = JID.from_string(self._client.get_own_jid().bare)

        self.pk = message.pk
        self.timestamp = message.timestamp
        self.local_timestamp = message.timestamp.astimezone()

        self.type = MessageType(message.type)

        self.contact = self._client.get_module("Contacts").get_contact(
            jid, groupchat=self.type == MessageType.GROUPCHAT
        )
        assert isinstance(
            self.contact, BareContact | GroupchatContact | GroupchatParticipant
        )

        self.add_css_class("search-view-row")
        self.add_css_class("opacity-0")

        self._ui = get_builder("search_view.ui", ["result_row_grid"])
        self.set_child(self._ui.result_row_grid)

        contact_name = self.contact.name
        if self.type == MessageType.GROUPCHAT:
            contact_name = message.resource or self.remote_jid.localpart
            assert contact_name is not None

        self._ui.row_name_label.set_text(contact_name)

        avatar = self._get_avatar(self.direction, contact_name)
        self._ui.row_avatar.set_pixel_size(AvatarSize.ROSTER)
        self._ui.row_avatar.set_from_paintable(avatar)

        format_string = app.settings.get("time_format")
        self._ui.row_time_label.set_text(self.local_timestamp.strftime(format_string))

        text = message.text
        assert text is not None

        message_widget = MessageWidget(self.account, selectable=False)
        message_widget.add_with_styling(text, nickname=contact_name)
        self._ui.result_row_grid.attach(message_widget, 1, 1, 2, 1)

        GLib.timeout_add(100, self.remove_css_class, "opacity-0")

    def do_unroot(self) -> None:
        Gtk.ListBoxRow.do_unroot(self)
        app.check_finalize(self)

    def _get_client(self, account_jid: str) -> Client:
        for client in app.get_clients():
            if client.is_own_jid(account_jid):
                return client

        raise ValueError("Unable to find account: %s" % account_jid)

    def _get_avatar(self, direction: ChatDirection, name: str) -> Gdk.Texture | None:
        scale = self.get_scale_factor()
        if isinstance(self.contact, GroupchatContact):
            contact = self.contact.get_resource(name)
            return contact.get_avatar(AvatarSize.ROSTER, scale, add_show=False)

        if direction == ChatDirection.OUTGOING:
            contact = self._client.get_module("Contacts").get_contact(
                self._client.get_own_jid().bare
            )
        else:
            contact = self.contact

        assert not isinstance(contact, GroupchatContact | ResourceContact)
        return contact.get_avatar(AvatarSize.ROSTER, scale, add_show=False)


@dataclass
class SearchFilterData:
    usernames: list[str] | None
    before: dt.datetime | None
    after: dt.datetime | None


class SearchFilters(Gtk.Expander, SignalManager):
    __gsignals__ = {
        "filter-changed": (GObject.SignalFlags.RUN_LAST, None, ()),
        "filter-activated": (
            GObject.SignalFlags.RUN_LAST,
            None,
            (),
        ),
    }

    def __init__(self) -> None:
        Gtk.Expander.__init__(self, label=_("Search Filters"))
        SignalManager.__init__(self)

        self._after: dt.datetime | None = None
        self._before: dt.datetime | None = None

        self._ui = get_builder(
            "search_view.ui", ["filter_date_selector_popover", "search_filters_grid"]
        )
        self.set_child(self._ui.search_filters_grid)

        self._connect(
            self._ui.filter_from_entry, "changed", self._on_filter_from_changed
        )
        self._connect(
            self._ui.filter_from_entry, "activate", self._on_from_entry_activated
        )
        self._connect(
            self._ui.filter_date_before_calendar, "day-selected", self._on_date_selected
        )
        self._connect(
            self._ui.filter_date_after_calendar, "day-selected", self._on_date_selected
        )

        self._connect(
            self._ui.filter_date_before_reset_button,
            "clicked",
            self._on_date_reset_clicked,
        )
        self._connect(
            self._ui.filter_date_after_reset_button,
            "clicked",
            self._on_date_reset_clicked,
        )

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Expander.do_unroot(self)
        app.check_finalize(self)

    def _on_filter_from_changed(self, _entry: Gtk.Entry) -> None:
        self._update_state()

    def _on_from_entry_activated(self, _entry: Gtk.Entry) -> None:
        self.emit("filter-activated")

    def _on_date_selected(self, calendar: Gtk.Calendar) -> None:
        g_datetime = calendar.get_date()
        datetime = dt.datetime(*g_datetime.get_ymd(), tzinfo=dt.UTC)
        date_format = app.settings.get("date_format")

        if self._ui.filter_before_button.get_active():
            self._before = datetime
            self._ui.filter_before_label.set_text(datetime.strftime(date_format))

        elif self._ui.filter_after_button.get_active():
            self._after = datetime
            self._ui.filter_after_label.set_text(datetime.strftime(date_format))

        else:
            raise ValueError

        self._update_state()
        self.emit("filter-activated")

    def _on_date_reset_clicked(self, _button: Gtk.Button) -> None:
        if self._ui.filter_before_button.get_active():
            self._before = None
            self._ui.filter_before_label.set_text("-")
            self._ui.filter_date_before_popover.popdown()

        elif self._ui.filter_after_button.get_active():
            self._after = None
            self._ui.filter_after_label.set_text("-")
            self._ui.filter_date_after_popover.popdown()

        else:
            raise ValueError

        self._update_state()
        self.emit("filter-activated")

    def _update_state(self) -> None:
        from_filter = self._ui.filter_from_entry.get_text()

        if any((from_filter, self._before, self._after)):
            self.set_label(_("Search Filters (Active)"))
        else:
            self.set_label(_("Search Filters"))

        self.emit("filter-changed")

    def reset(self) -> None:
        self._before = None
        self._after = None

        self._ui.filter_before_label.set_text("-")
        self._ui.filter_after_label.set_text("-")

        self._ui.filter_from_entry.set_text("")

        self.set_label(_("Search Filters"))

    def set_context(self, account: str | None, jid: JID | None) -> None:
        if account is None or jid is None:
            self._ui.filter_from_desc_label.set_visible(False)
            self._ui.filter_from_entry.set_visible(False)
            return

        client = app.get_client(account)
        contact = client.get_module("Contacts").get_contact(jid)
        visible = not isinstance(contact, BareContact)
        self._ui.filter_from_desc_label.set_visible(visible)
        self._ui.filter_from_entry.set_visible(visible)

    def get_filters(self) -> SearchFilterData:
        usernames: list[str] = []
        username = self._ui.filter_from_entry.get_text() or None
        if username is not None:
            usernames.append(username)

        return SearchFilterData(
            usernames=usernames or None,
            before=self._before,
            after=self._after,
        )



# Copyright (C) 2003-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2004 Vincent Hanquez <tab AT snarc.org>
# Copyright (C) 2005-2006 Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2009 Benjamin Richter <br AT waldteufel-online.net>
# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import cast

import gettext
import importlib.resources
import locale
import os
import sys
import unicodedata
from pathlib import Path

DOMAIN = 'gajim'


def init() -> None:
    _trans.init()


def get_default_lang() -> str:
    return _trans.get_default_lang()


def get_short_lang_code(lang: str | None = None) -> str:
    if lang is None:
        lang = _trans.get_default_lang()
    return lang[:2]


def is_rtl_text(text: str) -> bool:
    '''
    Determine paragraph writing direction according to
    http://www.unicode.org/reports/tr9/#The_Paragraph_Level
    '''
    for char in text:
        bidi = unicodedata.bidirectional(char)
        if bidi == 'L':
            return False
        if bidi in ('AL', 'R'):
            return True

    return False


def ngettext(s_sing: str, s_plural: str, n: int) -> str:
    return _trans.translation.ngettext(s_sing, s_plural, n)


class Translation:
    def __init__(self) -> None:
        self.translation = gettext.NullTranslations()
        self._default_lang = None

        self.install()

    def get_default_lang(self) -> str:
        assert self._default_lang is not None
        return self._default_lang

    @staticmethod
    def _get_win32_default_lang() -> str:
        if env_lang := os.environ.get('LANGUAGE'):
            return env_lang

        import ctypes
        windll = ctypes.windll.kernel32
        return locale.windows_locale[windll.GetUserDefaultUILanguage()]

    @staticmethod
    def _get_darwin_default_lang() -> str:
        from AppKit import NSLocale

        # FIXME: This returns a two letter language code (en, de, fr)
        # We need a way to get en_US, de_DE etc.
        return NSLocale.currentLocale().languageCode()

    def _get_default_lang(self) -> str:
        if sys.platform == 'win32':
            return self._get_win32_default_lang()

        if sys.platform == 'darwin':
            return self._get_darwin_default_lang()

        return locale.getlocale()[0] or 'en'

    def init(self) -> None:
        try:
            locale.setlocale(locale.LC_ALL, '')
        except locale.Error as error:
            print(error, file=sys.stderr)

        self._default_lang = self._get_default_lang()
        if sys.platform == 'win32':
            # Set the env var on Windows because gettext.find() uses it to
            # find the translation
            # Use LANGUAGE instead of LANG, LANG sets LC_ALL and thus
            # doesn't retain other region settings like LC_TIME
            os.environ['LANGUAGE'] = self._default_lang

        package_dir = cast(Path, importlib.resources.files('gajim'))
        locale_dir = package_dir / 'data' / 'locale'

        try:
            self.translation = gettext.translation(DOMAIN, locale_dir)
            if hasattr(locale, 'bindtextdomain'):
                locale.bindtextdomain(DOMAIN, locale_dir)
                locale.textdomain(DOMAIN)
        except OSError:
            pass

        self.install()

    def install(self) -> None:
        global _, g_, p_
        _ = self.translation.gettext
        g_ = self.translation.gettext
        p_ = self.translation.pgettext


_trans = Translation()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

import sys

from gi.repository import Adw
from gi.repository import Gtk

from gajim.common import app
from gajim.common import configpaths
from gajim.common import ged
from gajim.common.const import THRESHOLD_OPTIONS
from gajim.common.events import StyleChanged
from gajim.common.events import ThemeUpdate
from gajim.common.i18n import _
from gajim.common.setting_values import BoolSettings

# from gajim.common.util.av import AudioOutputManager
# from gajim.common.util.av import VideoInputManager
from gajim.common.util.av import AudioInputManager
from gajim.common.util.version import package_version

from gajim.gtk.alert import ConfirmationAlertDialog
from gajim.gtk.const import Setting
from gajim.gtk.const import SettingKind
from gajim.gtk.const import SettingType
from gajim.gtk.plugins import Plugins
from gajim.gtk.settings import DropDownSetting
from gajim.gtk.settings import GajimPreferencePage
from gajim.gtk.settings import GajimPreferencesGroup
from gajim.gtk.settings import SpinRange
from gajim.gtk.util.misc import open_file
from gajim.gtk.util.window import get_app_window
from gajim.gtk.util.window import open_window

# from gajim.gtk.video_preview import VideoPreview


class WindowBehaviourGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="window-behaviour",
            title=_("Behaviour of Application Window"),
        )

        main_window_on_startup_items = {
            "always": _("Always"),
            "never": _("Never"),
            "last_state": _("Restore last state"),
        }

        action_on_close_items = {
            "hide": _("Hide"),
            "minimize": _("Minimize"),
            "quit": _("Quit"),
        }

        settings = [
            Setting(
                SettingKind.DROPDOWN,
                _("Show on Startup"),
                SettingType.CONFIG,
                "show_main_window_on_startup",
                props={"data": main_window_on_startup_items},
                desc=_("Show window when starting Gajim"),
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Action on Close"),
                SettingType.CONFIG,
                "action_on_close",
                props={"data": action_on_close_items},
                desc=_("Action when closing Gajim’s window"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show in Taskbar"),
                SettingType.CONFIG,
                "show_in_taskbar",
                desc=_("Show window in the taskbar"),
                callback=self._on_show_in_taskbar,
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show Title Bar"),
                SettingType.CONFIG,
                "show_header_bar",
                desc=_("Show the window’s title bar"),
                enabled_func=lambda: sys.platform != "win32",
            ),
        ]

        for setting in settings:
            self.add_setting(setting)

    @staticmethod
    def _on_show_in_taskbar(value: bool, *args: Any) -> None:
        app.window.set_skip_taskbar_hint(not value)


class PluginSettingsGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="plugin-settings",
            title=_("Plugin Settings"),
        )

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Check for updates"),
                SettingType.CONFIG,
                "plugins_update_check",
                desc=_("Check for updates periodically"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Update automatically"),
                SettingType.CONFIG,
                "plugins_auto_update",
                desc=_("Update plugins automatically"),
                bind="plugins_update_check",
            ),
            Setting(
                SettingKind.SWITCH,
                _("Notify after update"),
                SettingType.CONFIG,
                "plugins_notify_after_update",
                desc=_("Notify me when the automatic update was successful"),
                bind="plugins_auto_update",
            ),
        ]

        for setting in settings:
            self.add_setting(setting)


class GeneralGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="general",
            title=_("General"),
        )

        speller_desc = None
        if not app.is_installed("SPELLING"):
            speller_desc = _("Needs libspelling to be installed")

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Close with Escape"),
                SettingType.CONFIG,
                "escape_key_closes",
                desc=_("Close a chat by pressing the Escape key"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show Send Message Button"),
                SettingType.CONFIG,
                "show_send_message_button",
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show Voice Message Button"),
                SettingType.CONFIG,
                "show_voice_message_button",
            ),
            Setting(
                SettingKind.SWITCH,
                _("Send Messages with Control+Enter"),
                SettingType.CONFIG,
                "send_on_ctrl_enter",
            ),
            Setting(
                SettingKind.SWITCH,
                _("Spell Checking"),
                SettingType.CONFIG,
                "use_speller",
                desc=speller_desc,
                enabled_func=self._speller_available,
            ),
            Setting(
                SettingKind.SWITCH,
                _("Emoji Shortcodes"),
                SettingType.CONFIG,
                "enable_emoji_shortcodes",
                desc=_("Show suggestions for shortcodes, e.g. :+1:"),
            ),
        ]

        for setting in settings:
            self.add_setting(setting)

    @staticmethod
    def _speller_available() -> bool:
        return app.is_installed("SPELLING")


class ChatsGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="chats",
            title=_("Chats"),
        )

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Message Receipts (✔)"),
                SettingType.CONFIG,
                "positive_184_ack",
                desc=_("Add a checkmark to received messages"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show Status Changes"),
                SettingType.CONFIG,
                "print_status_in_chats",
                desc=_('For example: "Julia is now online"'),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show Link Preview"),
                SettingType.CONFIG,
                "enable_link_preview_default",
                desc=_("Default for new chats"),
                callback=app.window.reload_view,
            ),
        ]

        for setting in settings:
            self.add_setting(setting)


class GroupChatsGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="group-chats",
            title=_("Group Chats"),
        )

        settings = [
            Setting(
                SettingKind.DROPDOWN,
                _("Default Sync Threshold"),
                SettingType.CONFIG,
                "gc_sync_threshold_public_default",
                desc=_("Default for new public group chats"),
                props={"data": THRESHOLD_OPTIONS},
            ),
            Setting(
                SettingKind.SWITCH,
                _("Direct Messages"),
                SettingType.CONFIG,
                "muc_prefer_direct_msg",
                desc=_("Prefer direct messages in private group chats "),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Sort Participant List by Status"),
                SettingType.CONFIG,
                "sort_by_show_in_muc",
                callback=self._on_sort_by_show_in_muc,
            ),
            Setting(
                SettingKind.SWITCH,
                _("Status Messages in Participants List"),
                SettingType.CONFIG,
                "show_status_msgs_in_roster",
                callback=self._on_show_status_in_roster,
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show Subject on Join"),
                SettingType.CONFIG,
                "show_subject_on_join",
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show Joined / Left"),
                SettingType.CONFIG,
                "gc_print_join_left_default",
                desc=_("Default for new group chats"),
                props={
                    "button-text": _("Reset"),
                    "button-tooltip": _(
                        "Reset all group chats to the current default value"
                    ),
                    "button-style": "destructive-action",
                    "button-callback": self._reset_join_left,
                },
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show Status Changes"),
                SettingType.CONFIG,
                "gc_print_status_default",
                desc=_("Default for new group chats"),
                props={
                    "button-text": _("Reset"),
                    "button-tooltip": _(
                        "Reset all group chats to the current default value"
                    ),
                    "button-style": "destructive-action",
                    "button-callback": self._reset_print_status,
                },
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show Link Preview"),
                SettingType.CONFIG,
                "gc_enable_link_preview_default",
                desc=_("Default for new group chats"),
                callback=app.window.reload_view,
            ),
        ]

        for setting in settings:
            self.add_setting(setting)

    @staticmethod
    def _on_sort_by_show_in_muc(_value: bool, *args: Any) -> None:
        roster = app.window.get_control().get_group_chat_roster()
        roster.invalidate_sort()

    @staticmethod
    def _on_show_status_in_roster(_value: bool, *args: Any) -> None:
        roster = app.window.get_control().get_group_chat_roster()
        roster.invalidate_sort()

    @staticmethod
    def _reset_join_left(button: Gtk.Button) -> None:
        button.set_sensitive(False)
        app.settings.set_group_chat_settings("print_join_left", None, None)

    @staticmethod
    def _reset_print_status(button: Gtk.Button) -> None:
        button.set_sensitive(False)
        app.settings.set_group_chat_settings("print_status", None, None)


class FilePreviewGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="file-preview",
            title=_("File Preview"),
        )

        sizes = {
            0: _("No automatic preview"),
            262144: "256 KiB",
            524288: "512 KiB",
            1048576: "1 MiB",
            5242880: "5 MiB",
            10485760: "10 MiB",
            26214400: "25 MiB",
        }

        settings = [
            Setting(
                SettingKind.DROPDOWN,
                _("File Size Limit"),
                SettingType.CONFIG,
                "preview_max_file_size",
                desc=_("Maximum file size for preview downloads"),
                props={"data": sizes},
            ),
            Setting(
                SettingKind.SPIN,
                _("Preview Size"),
                SettingType.CONFIG,
                "preview_size",
                desc=_("Size of preview images in pixels"),
                props={"range_": SpinRange(100, 1000, 50, 0)},
            ),
            Setting(
                SettingKind.SWITCH,
                _("Preview in Public Group Chats"),
                SettingType.CONFIG,
                "preview_anonymous_muc",
                desc=_(
                    "Show previews automatically in public "
                    "group chats (may disclose your data)"
                ),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Preview all Image URLs"),
                SettingType.CONFIG,
                "preview_allow_all_images",
                desc=_("Show previews for any URLs containing images (may be unsafe)"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("HTTPS Verification"),
                SettingType.CONFIG,
                "preview_verify_https",
                desc=_(
                    "Whether to check for a valid certificate before "
                    "downloading (not safe to disable)"
                ),
            ),
        ]

        for setting in settings:
            self.add_setting(setting)


class SystemNotificationsGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="system-notifications",
            title=_("System"),
        )

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("System Tray Icon"),
                SettingType.CONFIG,
                "show_trayicon",
            ),
            Setting(
                SettingKind.SWITCH,
                _("System Tray Notifications"),
                SettingType.CONFIG,
                "trayicon_notification_on_events",
                desc=_("Show notifications on the system tray icon"),
                bind="show_trayicon",
            ),
            Setting(
                SettingKind.SWITCH,
                _("System Notifications"),
                SettingType.CONFIG,
                "show_notifications",
            ),
            Setting(
                SettingKind.SWITCH,
                _("System Notifications When Away"),
                SettingType.CONFIG,
                "show_notifications_away",
                desc=_("Show notifications even if you are Away, Busy, etc."),
                bind="show_notifications",
            ),
        ]

        for setting in settings:
            self.add_setting(setting)


class NotificationsChatsGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="notifications-chats",
            title=_("Chats"),
        )

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Reactions"),
                SettingType.CONFIG,
                "notify_on_reaction_default",
                desc=_("Notify me when someone reacts to my message in a chat"),
            ),
        ]

        for setting in settings:
            self.add_setting(setting)


class NotificationsGroupChatsGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="notifications-group-chats",
            title=_("Group Chats"),
        )

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Reactions"),
                SettingType.CONFIG,
                "gc_notify_on_reaction_default",
                desc=_("Notify me when someone reacts to my message in a group chat"),
            ),
        ]

        for setting in settings:
            self.add_setting(setting)


class SoundsGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="sounds",
            title=_("Sounds"),
        )

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Play Sounds"),
                SettingType.CONFIG,
                "sounds_on",
                desc=_("Play sounds to notify about events"),
                props={
                    "button-icon-name": "lucide-settings-symbolic",
                    "button-callback": self._on_manage_sounds,
                },
            ),
            Setting(
                SettingKind.SWITCH,
                _("Sounds When Away"),
                SettingType.CONFIG,
                "sounddnd",
                desc=_("Play sounds even when you are Away, Busy, etc."),
                bind="sounds_on",
            ),
        ]

        for setting in settings:
            self.add_setting(setting)

    @staticmethod
    def _on_manage_sounds(_button: Gtk.Button) -> None:
        window = get_app_window("Preferences")
        assert window is not None
        open_window("ManageSounds", transient_for=window)


class StatusMessageGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="status-message",
            title=_("Ask For Status Message on…"),
        )

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Sign In"),
                SettingType.CONFIG,
                "ask_online_status",
            ),
        ]

        for setting in settings:
            self.add_setting(setting)


class AutomaticStatusGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="automatic-status",
            title=_("Automatic Status Change"),
        )

        settings = [
            Setting(
                SettingKind.SUBPAGE,
                _("Auto Away"),
                SettingType.CONFIG,
                "autoaway",
                desc=_('Change your status to "Away" after a certain amount of time'),
                props={"subpage": "auto-away"},
            ),
            Setting(
                SettingKind.SUBPAGE,
                _("Auto Not Available"),
                SettingType.CONFIG,
                "autoxa",
                desc=_(
                    'Change your status to "Not Available" after a '
                    "certain amount of time"
                ),
                props={"subpage": "auto-extended-away"},
            ),
        ]

        for setting in settings:
            self.add_setting(setting)

    @staticmethod
    def _get_auto_away() -> bool:
        return app.settings.get("autoaway")

    @staticmethod
    def _get_auto_xa() -> bool:
        return app.settings.get("autoxa")


class AutoAwayGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(self, key="auto-away")

        settings = [
            Setting(SettingKind.SWITCH, _("Auto Away"), SettingType.CONFIG, "autoaway"),
            Setting(
                SettingKind.SPIN,
                _("Time Until Away"),
                SettingType.CONFIG,
                "autoawaytime",
                desc=_("Minutes until your status gets changed"),
                props={"range_": SpinRange(1, 720, 1, 0)},
                bind="autoaway",
            ),
            Setting(
                SettingKind.ENTRY,
                _("Status Message"),
                SettingType.CONFIG,
                "autoaway_message",
                bind="autoaway",
            ),
        ]

        for setting in settings:
            self.add_setting(setting)


class AutoExtendedAwayGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(self, key="auto-extended-away")

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Auto Not Available"),
                SettingType.CONFIG,
                "autoxa",
            ),
            Setting(
                SettingKind.SPIN,
                _("Time Until Not Available"),
                SettingType.CONFIG,
                "autoxatime",
                desc=_("Minutes until your status gets changed"),
                props={"range_": SpinRange(1, 720, 1, 0)},
                bind="autoxa",
            ),
            Setting(
                SettingKind.ENTRY,
                _("Status Message"),
                SettingType.CONFIG,
                "autoxa_message",
                bind="autoxa",
            ),
        ]

        for setting in settings:
            self.add_setting(setting)


class ThemesGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="themes",
            title=_("Themes"),
        )

        theme_items = self._get_theme_items()

        dark_theme_items = {
            0: _("Disabled"),
            1: _("Enabled"),
            2: _("System"),
        }

        settings = [
            Setting(
                SettingKind.SPIN,
                _("User Interface Font Size"),
                SettingType.CONFIG,
                "app_font_size",
                props={"range_": SpinRange(0.5, 3.0, 0.10, 2)},
                callback=self._on_app_font_size_changed,
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Dark Theme"),
                SettingType.CONFIG,
                "dark_theme",
                props={"data": dark_theme_items},
                callback=self._on_dark_theme,
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Theme"),
                SettingType.CONFIG,
                "roster_theme",
                name="roster_theme",
                props={
                    "data": theme_items,
                    "button-icon-name": "lucide-settings-symbolic",
                    "button-callback": self._on_edit_themes,
                },
                callback=self._on_theme_changed,
            ),
        ]

        for setting in settings:
            self.add_setting(setting)

        app.ged.register_event_handler("theme-update", ged.GUI1, self._on_theme_update)

    def do_unroot(self) -> None:
        app.ged.remove_event_handler("theme-update", ged.GUI1, self._on_theme_update)
        GajimPreferencesGroup.do_unroot(self)

    @staticmethod
    def _on_app_font_size_changed(_value: float, *args: Any) -> None:
        app.css_config.apply_app_font_size()

    @staticmethod
    def _get_theme_items() -> list[str]:
        return ["default", *app.css_config.themes]

    def _on_theme_update(self, *args: Any) -> None:
        dropdown_row = cast(DropDownSetting, self.get_setting("roster_theme"))
        dropdown_row.update_entries(self._get_theme_items())
        dropdown_row.select_key(app.settings.get("roster_theme"))

    def _on_edit_themes(self, _button: Gtk.Button) -> None:
        window = get_app_window("Preferences")
        assert window is not None
        open_window("Themes", transient=window)

    def _on_theme_changed(self, value: str, *args: Any) -> None:
        app.ged.remove_event_handler("theme-update", ged.GUI1, self._on_theme_update)

        app.css_config.change_theme(value)
        app.ged.raise_event(ThemeUpdate())

        app.ged.register_event_handler("theme-update", ged.GUI1, self._on_theme_update)

    @staticmethod
    def _on_dark_theme(value: str, *args: Any) -> None:
        app.css_config.set_dark_theme(int(value))
        app.css_config.reload_css()
        app.ged.raise_event(StyleChanged())


# class ServerGroup(GajimPreferencesGroup):
#     def __init__(self) -> None:
#         GajimPreferencesGroup.__init__(self, key="server")

#         settings = [
#             Setting(
#                 SettingKind.SUBPAGE,
#                 _("Use STUN Server"),
#                 SettingType.DIALOG,
#                 desc=_("Helps to establish calls through firewalls"),
#                 props={"subpage": "stun"},
#             ),
#         ]

#         for setting in settings:
#             self.add_setting(setting)

#         self.set_sensitive(app.is_installed("AV"))

# class StunServerGroup(GajimPreferencesGroup):
#     def __init__(self) -> None:
#         GajimPreferencesGroup.__init__(self, key="stun")

#         settings = [
#             Setting(
#                 SettingKind.SWITCH,
#                 _("Use STUN Server"),
#                 SettingType.CONFIG,
#                 "use_stun_server",
#             ),
#             Setting(
#                 SettingKind.ENTRY,
#                 _("STUN Server"),
#                 SettingType.CONFIG,
#                 "stun_server",
#                 bind="use_stun_server",
#             ),
#         ]

#         for setting in settings:
#             self.add_setting(setting)


class AudioGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="audio",
            title=_("Audio"),
        )

        deps_installed = app.is_installed("GST")

        audio_input_devices = {}
        # audio_output_devices = {}
        if deps_installed:
            audio_input_devices = AudioInputManager().get_devices()
            # audio_output_devices = AudioOutputManager().get_devices()

        audio_input_items = self._create_av_combo_items(audio_input_devices)
        # audio_output_items = self._create_av_combo_items(audio_output_devices)

        settings = [
            Setting(
                SettingKind.DROPDOWN,
                _("Audio Input Device"),
                SettingType.CONFIG,
                "audio_input_device",
                desc=_("Select your audio input (e.g. microphone)"),
                props={"data": audio_input_items},
            ),
            # Setting(
            #     SettingKind.DROPDOWN,
            #     _("Audio Output Device"),
            #     SettingType.CONFIG,
            #     "audio_output_device",
            #     desc=_("Select an audio output (e.g. speakers, headphones)"),
            #     props={"data": audio_output_items},
            # ),
        ]

        for setting in settings:
            self.add_setting(setting)

        self.set_sensitive(deps_installed)

    @staticmethod
    def _create_av_combo_items(items_dict: dict[str, str]) -> dict[str, str]:
        items = enumerate(
            sorted(
                items_dict.items(),
                key=lambda x: "" if x[1].startswith("auto") else x[0].lower(),
            )
        )
        combo_items: dict[str, str] = {}
        for _index, (name, value) in items:
            combo_items[value] = name
        return combo_items


# class VideoGroup(GajimPreferencesGroup):
#     def __init__(self) -> None:
#         GajimPreferencesGroup.__init__(self)

#         deps_installed = app.is_installed("AV")

#         video_input_devices = {}
#         if deps_installed:
#             video_input_devices = VideoInputManager().get_devices()

#         video_input_items = self._create_av_combo_items(video_input_devices)

#         video_framerates = {
#             "": _("Default"),
#             "15/1": "15 fps",
#             "10/1": "10 fps",
#             "5/1": "5 fps",
#             "5/2": "2.5 fps",
#         }

#         video_sizes = {
#             "": _("Default"),
#             "800x600": "800x600",
#             "640x480": "640x480",
#             "320x240": "320x240",
#         }

#         settings = [
#             Setting(
#                 SettingKind.DROPDOWN,
#                 _("Video Input Device"),
#                 SettingType.CONFIG,
#                 "video_input_device",
#                 props={"data": video_input_items},
#                 desc=_("Select your video input device (e.g. webcam, screen capture)"),  # noqa: E501
#                 callback=self._on_video_input_changed,
#             ),
#             Setting(
#                 SettingKind.DROPDOWN,
#                 _("Video Framerate"),
#                 SettingType.CONFIG,
#                 "video_framerate",
#                 props={"data": video_framerates},
#             ),
#             Setting(
#                 SettingKind.DROPDOWN,
#                 _("Video Resolution"),
#                 SettingType.CONFIG,
#                 "video_size",
#                 props={"data": video_sizes},
#             ),
#             Setting(
#                 SettingKind.SWITCH,
#                 _("Show My Video Stream"),
#                 SettingType.CONFIG,
#                 "video_see_self",
#                 desc=_("Show your own video stream in calls"),
#             ),
#             Setting(
#                 SettingKind.SWITCH,
#                 _("Live Preview"),
#                 SettingType.VALUE,
#                 desc=_("Show a live preview to test your video source"),
#                 callback=self._toggle_live_preview,
#             ),
#         ]

#         for setting in settings:
#             self.add_setting(setting)

#         self.set_sensitive(deps_installed)

#     @staticmethod
#     def _on_video_input_changed(_value: str, *args: Any) -> None:
#         window = get_app_window("Preferences")
#         assert window is not None
#         preview = window.get_video_preview()
#         if preview is None or not preview.is_active:
#             # changed signal gets triggered when we fill the dropdown
#             return
#         preview.refresh()

#     @staticmethod
#     def _toggle_live_preview(value: bool, *args: Any) -> None:
#         window = get_app_window("Preferences")
#         assert window is not None
#         preview = window.get_video_preview()
#         if preview is not None:
#             preview.toggle_preview(value)

#     @staticmethod
#     def _create_av_combo_items(items_dict: dict[str, str]) -> dict[str, str]:
#         items = enumerate(
#             sorted(
#                 items_dict.items(),
#                 key=lambda x: "" if x[1].startswith("auto") else x[0].lower(),
#             )
#         )
#         combo_items: dict[str, str] = {}
#         for _index, (name, value) in items:
#             combo_items[value] = name
#         return combo_items


class MiscellaneousGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="miscellaneous",
            title=_("Miscellaneous"),
        )

        self._hints_list: list[BoolSettings] = [
            "show_help_start_chat",
        ]

        settings = [
            Setting(
                SettingKind.DROPDOWN,
                _("Global Proxy"),
                SettingType.CONFIG,
                "global_proxy",
                name="global_proxy",
                props={
                    "data": self._get_proxies(),
                    "button-icon-name": "lucide-settings-symbolic",
                    "button-callback": self._on_proxy_edit,
                },
            ),
            Setting(
                SettingKind.SWITCH,
                _("Use System Keyring"),
                SettingType.CONFIG,
                "use_keyring",
                desc=_("Use your system’s keyring to store passwords"),
            ),
        ]

        if sys.platform not in ("win32", "darwin") and package_version(
            "keyring>=23.8.1"
        ):
            settings.append(
                Setting(
                    SettingKind.SWITCH,
                    _("Enable KeepassXC Integration"),
                    SettingType.CONFIG,
                    "enable_keepassxc_integration",
                )
            )

        if sys.platform in ("win32", "darwin"):
            settings.append(
                Setting(
                    SettingKind.SWITCH,
                    _("Check For Updates"),
                    SettingType.CONFIG,
                    "check_for_update",
                    desc=_("Check for Gajim updates periodically"),
                )
            )

        settings.append(
            Setting(
                SettingKind.GENERIC,
                _("Reset Hints"),
                SettingType.VALUE,
                None,
                desc=_("This will reset all introduction hints"),
                props={
                    "button-text": _("Reset"),
                    "button-style": "destructive-action",
                    "button-callback": self._on_reset_hints,
                    "button-sensitive": self._check_hints_reset(),
                },
            ),
        )

        settings.append(
            Setting(
                SettingKind.GENERIC,
                _("Purge Chat History"),
                SettingType.VALUE,
                None,
                desc=_("This will purge all locally stored chat history"),
                props={
                    "button-text": _("Purge"),
                    "button-style": "destructive-action",
                    "button-callback": self._on_purge_history_clicked,
                },
            ),
        )

        for setting in settings:
            self.add_setting(setting)

    @staticmethod
    def _get_proxies() -> dict[str, str]:
        proxies = {"": _("System")}
        proxies.update({proxy: proxy for proxy in app.settings.get_proxies()})
        proxies["no-proxy"] = _("No Proxy")
        return proxies

    @staticmethod
    def _on_proxy_edit(*args: Any) -> None:
        window = get_app_window("Preferences")
        assert window is not None
        open_window("ManageProxies", transient_for=window)

    def update_proxy_list(self) -> None:
        dropdown_row = cast(DropDownSetting, self.get_setting("global_proxy"))
        dropdown_row.update_entries(self._get_proxies())

    def _check_hints_reset(self) -> bool:
        return any(app.settings.get(hint) is False for hint in self._hints_list)

    def _on_reset_hints(self, button: Gtk.Button) -> None:
        for hint in self._hints_list:
            app.settings.set(hint, True)
        button.set_sensitive(False)

    @staticmethod
    def _on_purge_history_clicked(button: Gtk.Button) -> None:
        def _on_response() -> None:
            button.set_sensitive(False)
            app.storage.archive.remove_all_history()
            app.app.start_shutdown()

        ConfirmationAlertDialog(
            _("Purge all Chat History?"),
            _(
                "Do you really want to remove all chat messages from Gajim?\n"
                "Warning: This can’t be undone!\n"
                "Gajim will quit afterwards."
            ),
            confirm_label=_("_Purge"),
            appearance="destructive",
            callback=_on_response,
        )


class AdvancedGroup(GajimPreferencesGroup):
    def __init__(self) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="advanced",
            title=_("Advanced Settings"),
        )

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Debug Logging"),
                SettingType.VALUE,
                app.get_debug_mode(),
                desc=_("Restart Gajim to start debug logging"),
                props={
                    "button-icon-name": "lucide-folder-symbolic",
                    "button-callback": self._on_open_debug_logs,
                },
                callback=self._on_debug_logging,
            ),
            Setting(
                SettingKind.GENERIC,
                _("Debug Console"),
                SettingType.VALUE,
                None,
                props={
                    "button-text": _("Open"),
                    "button-callback": self._on_debug,
                },
            ),
            Setting(
                SettingKind.SWITCH,
                _("D-Bus Interface"),
                SettingType.CONFIG,
                "remote_control",
                desc=_(
                    "Allow Gajim to broadcast useful information via "
                    "D-Bus. It also allows other applications to "
                    "control Gajim remotely."
                ),
            ),
            Setting(
                SettingKind.GENERIC,
                _("Advanced Configuration Editor"),
                SettingType.VALUE,
                None,
                desc=_("Please use these settings with caution!"),
                props={
                    "button-text": _("Open"),
                    "button-callback": self._on_advanced_config_editor,
                },
            ),
        ]

        for setting in settings:
            self.add_setting(setting)

    @staticmethod
    def _on_debug_logging(value: bool, *args: Any) -> None:
        app.set_debug_mode(value)

    @staticmethod
    def _on_open_debug_logs(*args: Any) -> None:
        open_file(configpaths.get("DEBUG"))

    @staticmethod
    def _on_advanced_config_editor(*args: Any) -> None:
        window = get_app_window("Preferences")
        assert window is not None
        open_window("AdvancedConfig", transient_for=window)

    @staticmethod
    def _on_debug(*args: Any) -> None:
        open_window("DebugConsoleWindow")


class GeneralPage(GajimPreferencePage):
    key = "general"
    icon_name = "lucide-app-window-symbolic"
    label = _("General")

    def __init__(self) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("General"),
            groups=[WindowBehaviourGroup],
        )


class ChatsPage(GajimPreferencePage):
    key = "chats"
    icon_name = "lucide-message-circle-symbolic"
    label = _("Chats")

    def __init__(self) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Chats"),
            groups=[
                GeneralGroup,
                ChatsGroup,
                GroupChatsGroup,
                FilePreviewGroup,
            ],
        )


class VisualNotificationsPage(GajimPreferencePage):
    key = "system-notifications"
    icon_name = "lucide-bell-symbolic"
    label = _("Notifications")

    def __init__(self) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Notifications"),
            groups=[
                SystemNotificationsGroup,
                NotificationsChatsGroup,
                NotificationsGroupChatsGroup,
                SoundsGroup,
            ],
        )


class StatusPage(GajimPreferencePage):
    key = "status"
    icon_name = "lucide-megaphone-symbolic"
    label = _("Status")

    def __init__(self) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Status"),
            groups=[
                StatusMessageGroup,
                AutomaticStatusGroup,
            ],
        )


class StylePage(GajimPreferencePage):
    key = "style"
    icon_name = "lucide-palette-symbolic"
    label = _("Style")

    def __init__(self) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Style"),
            groups=[ThemesGroup],
        )


class AudioVideoPage(GajimPreferencePage):
    key = "audio-video"
    icon_name = "lucide-mic-symbolic"
    label = _("Audio")

    def __init__(self) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Audio"),
            groups=[
                # ServerGroup,
                AudioGroup,
                # VideoGroup,
            ],
        )

        banner = Adw.Banner(
            revealed=bool(not app.is_installed("GST")),
            title=_("Missing dependencies for audio support"),
            button_label=_("Features"),
            action_name="app.features",
        )
        self._pref_page.set_banner(banner)


class PluginsPage(GajimPreferencePage):
    key = "plugins"
    icon_name = "lucide-package-symbolic"
    label = _("Plugins")

    def __init__(self) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Plugins"),
            groups=[
                PluginSettingsGroup,
                Plugins,
            ],
        )


class AdvancedPage(GajimPreferencePage):
    key = "advanced"
    icon_name = "lucide-wrench-symbolic"
    label = _("Advanced")

    def __init__(self) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Advanced"),
            groups=[
                MiscellaneousGroup,
                AdvancedGroup,
            ],
        )


class AutoAwayPage(GajimPreferencePage):
    key = "auto-away"

    def __init__(self) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Auto Away"),
            groups=[AutoAwayGroup],
        )


class AutoExtendedAwayPage(GajimPreferencePage):
    key = "auto-extended-away"

    def __init__(self) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Auto Not Available"),
            groups=[AutoExtendedAwayGroup],
        )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0030: Service Discovery

from __future__ import annotations

from typing import Any

from collections.abc import Generator

import nbxmpp
from nbxmpp.errors import is_error
from nbxmpp.errors import StanzaError
from nbxmpp.modules.muc.util import MucInfoResult
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import IqProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import types
from gajim.common.events import MucDiscoUpdate
from gajim.common.events import ServerDiscoReceived
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.util import as_task


class Discovery(BaseModule):

    _nbxmpp_extends = 'Discovery'
    _nbxmpp_methods = [
        'disco_info',
        'disco_items',
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(name='iq',
                          callback=self._answer_disco_info,
                          typ='get',
                          ns=Namespace.DISCO_INFO),
            StanzaHandler(name='iq',
                          callback=self._answer_disco_items,
                          typ='get',
                          ns=Namespace.DISCO_ITEMS),
        ]

        self._account_info: DiscoInfo | None = None
        self._server_info: DiscoInfo | None = None

    @property
    def account_info(self) -> DiscoInfo | None:
        return self._account_info

    @property
    def server_info(self) -> DiscoInfo | None:
        return self._server_info

    def discover_server_items(self) -> None:
        server = self._con.get_own_jid().domain
        self.disco_items(server, callback=self._server_items_received)

    def _server_items_received(self, task: Task) -> None:
        try:
            result = task.finish()
        except StanzaError as error:
            self._log.warning('Server disco failed')
            self._log.error(error)
            return

        self._log.info('Server items received')
        self._log.debug(result)
        for item in result.items:
            if item.node is not None:
                # Only disco components
                continue
            self.disco_info(item.jid, callback=self._server_items_info_received)

    def _server_items_info_received(self, task: Task) -> None:
        try:
            result = task.finish()
        except StanzaError as error:
            self._log.warning('Server item disco info failed')
            self._log.warning(error)
            return

        self._log.info('Server item info received: %s', result.jid)
        self._parse_transports(result)
        try:
            self._con.get_module('MUC').pass_disco(result)
            self._con.get_module('HTTPUpload').pass_disco(result)
            self._con.get_module('Bytestream').pass_disco(result)
        except nbxmpp.NodeProcessed:
            pass

        app.ged.raise_event(ServerDiscoReceived(account=self._account))

    def discover_account_info(self) -> None:
        own_jid = self._con.get_own_jid().bare
        self.disco_info(own_jid, callback=self._account_info_received)

    def _account_info_received(self, task: Task) -> None:
        try:
            result = task.finish()
        except StanzaError as error:
            self._log.warning('Account disco info failed')
            self._log.warning(error)
            return

        self._log.info('Account info received: %s', result.jid)

        self._account_info = result

        self._con.get_module('MAM').pass_disco(result)
        self._con.get_module('PEP').pass_disco(result)
        self._con.get_module('PubSub').pass_disco(result)
        self._con.get_module('Bookmarks').pass_disco(result)
        self._con.get_module('VCardAvatars').pass_disco(result)
        self._con.get_module('ChatMarkers').pass_disco(result)

        self._con.get_module('Caps').update_caps()

    def discover_server_info(self) -> None:
        # Calling this method starts the connect_maschine()
        server = self._con.get_own_jid().domain
        self.disco_info(server, callback=self._server_info_received)

    def _server_info_received(self, task: Task) -> None:
        try:
            result = task.finish()
        except StanzaError as error:
            self._log.error('Server disco info failed')
            self._log.error(error)
            return

        self._log.info('Server info received: %s', result.jid)

        self._server_info = result

        self._con.get_module('SecLabels').pass_disco(result)
        self._con.get_module('Blocking').pass_disco(result)
        self._con.get_module('VCardTemp').pass_disco(result)
        self._con.get_module('Carbons').pass_disco(result)
        self._con.get_module('HTTPUpload').pass_disco(result)
        self._con.get_module('Register').pass_disco(result)

        self._con.connect_machine(restart=True)

    def _parse_transports(self, info: DiscoInfo) -> None:
        for identity in info.identities:
            if identity.category != 'gateway':
                continue

            self._log.info('Found transport: %s %s %s',
                           info.jid, identity.category, identity.type)

            for child in self._con.get_module(
                    'Contacts').get_contacts_with_domain(info.jid.domain):
                if not isinstance(child, BareContact | GroupchatContact):
                    continue
                child.update_gateway_type(identity.type)

            jid = str(info.jid)
            if jid not in app.transport_type:
                app.transport_type[jid] = identity.type

            if identity.type in self._con.available_transports:
                self._con.available_transports[identity.type].append(jid)
            else:
                self._con.available_transports[identity.type] = [jid]

    def _answer_disco_items(self,
                            _con: types.NBXMPPClient,
                            stanza: Iq,
                            _properties: IqProperties
                            ) -> None:
        from_ = stanza.getFrom()
        self._log.info('Answer disco items to %s', from_)

        node = stanza.getTagAttr('query', 'node')
        if node is None:
            result = stanza.buildReply('result')
            self._con.connection.send(result)
            raise nbxmpp.NodeProcessed

    def _answer_disco_info(self,
                           _con: types.NBXMPPClient,
                           stanza: Iq,
                           _properties: IqProperties
                           ) -> None:
        from_ = stanza.getFrom()
        self._log.info('Answer disco info %s', from_)
        if str(from_).startswith('echo.'):
            # Service that echos all stanzas, ignore it
            raise nbxmpp.NodeProcessed

    @as_task
    def disco_muc(self,
                  jid: JID | str,
                  request_vcard: bool = False,
                  allow_redirect: bool = False
                  ) -> Generator[Any, Any]:

        _task = yield  # noqa: F841

        self._log.info('Request MUC info for %s', jid)

        result = yield self._nbxmpp('MUC').request_info(
            jid,
            request_vcard=request_vcard,
            allow_redirect=allow_redirect)

        if is_error(result):
            raise result

        assert isinstance(result, MucInfoResult)
        if result.redirected:
            self._log.info('MUC info received after redirect: %s -> %s',
                           jid, result.info.jid)
        else:
            self._log.info('MUC info received: %s', result.info.jid)

        assert result.info.jid is not None
        app.storage.cache.set_last_disco_info(result.info.jid, result.info)

        if result.vcard is not None:
            avatar, avatar_sha = result.vcard.get_avatar()
            if avatar is not None:
                if not app.app.avatar_storage.avatar_exists(avatar_sha):
                    app.app.avatar_storage.save_avatar(avatar)

                app.storage.cache.set_muc(
                    self._account, result.info.jid, 'avatar', avatar_sha)
                app.app.avatar_storage.invalidate_cache(result.info.jid)

        self._con.get_module('VCardAvatars').muc_disco_info_update(result.info)
        app.ged.raise_event(MucDiscoUpdate(
            account=self._account,
            jid=result.info.jid))

        contact = self._con.get_module('Contacts').get_contact(
            result.info.jid, groupchat=True)
        contact.notify('disco-info-update')

        yield result

    @as_task
    def disco_contact(self, contact: types.ContactT) -> Generator[Any, Any]:
        _task = yield  # noqa: F841

        result = yield self.disco_info(contact.jid)
        if is_error(result):
            raise result

        self._log.info('Disco Info received: %s', contact.jid)

        assert isinstance(result, DiscoInfo)
        assert result.jid is not None
        app.storage.cache.set_last_disco_info(result.jid,
                                              result,
                                              cache_only=True)

        contact = self._con.get_module('Contacts').get_contact(result.jid)
        contact.notify('caps-update')



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gajim.common import types
from gajim.common.const import Draft
from gajim.common.helpers import Observable


class DraftStorage(Observable):
    def __init__(self) -> None:
        Observable.__init__(self)

        self._drafts: dict[types.ChatContactT, Draft] = {}

    def set(
        self,
        contact: types.ChatContactT,
        draft: Draft | None,
    ) -> None:
        if draft is None:
            self.remove(contact)
            return

        self._drafts[contact] = draft
        self.notify("draft-update", contact, draft)

    def get(self, contact: types.ChatContactT) -> Draft | None:
        return self._drafts.get(contact)

    def remove(self, contact: types.ChatContactT) -> None:
        self._drafts.pop(contact, None)
        self.notify("draft-update", contact, None)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from datetime import datetime

from gi.repository import Gtk

from gajim.common.const import AvatarSize
from gajim.common.i18n import _
from gajim.common.util.datetime import utc_now
from gajim.common.util.text import process_non_spacing_marks

from gajim.gtk.conversation.rows.base import BaseRow
from gajim.gtk.conversation.rows.widgets import DateTimeLabel
from gajim.gtk.conversation.rows.widgets import SimpleLabel


class MUCJoinLeft(BaseRow):
    def __init__(
        self,
        type_: str,
        account: str,
        nick: str,
        reason: str | None = None,
        error: bool = False,
        timestamp: datetime | None = None,
    ) -> None:
        BaseRow.__init__(self, account)

        self.type = type_
        if timestamp is None:
            timestamp = utc_now()
        self.timestamp = timestamp.astimezone()
        self.db_timestamp = timestamp.timestamp()

        avatar_placeholder = Gtk.Box()
        avatar_placeholder.set_size_request(AvatarSize.ROSTER, -1)
        self.grid.attach(avatar_placeholder, 0, 0, 1, 1)

        if type_ == "muc-user-joined":
            text = self._make_join_message(nick)
            icon_name = "lucide-log-in-symbolic"
            icon_class = "gajim-user-connected"
        else:
            text = self._make_left_message(nick, reason, error)
            icon_name = "lucide-log-out-symbolic"
            icon_class = "gajim-user-disconnected"

        icon = Gtk.Image.new_from_icon_name(icon_name)
        icon.add_css_class(icon_class)
        self.grid.attach(icon, 1, 0, 1, 1)

        self._label = SimpleLabel()
        self._label.set_text(process_non_spacing_marks(text))
        self._label.add_css_class("gajim-status-message")
        self.grid.attach(self._label, 2, 0, 1, 1)

        timestamp_widget = DateTimeLabel(self.timestamp)
        timestamp_widget.set_halign(Gtk.Align.START)
        timestamp_widget.set_valign(Gtk.Align.FILL)
        self.grid.attach(timestamp_widget, 3, 0, 1, 1)

    def do_unroot(self) -> None:
        BaseRow.do_unroot(self)

    @staticmethod
    def _make_left_message(nick: str, reason: str | None, error: bool) -> str:
        reason = "" if reason is None else f": {reason}"

        if error:
            # Group Chat: User was kicked because of an server error: reason
            message = _("{nick} has left due to an error{reason}").format(
                nick=nick, reason=reason
            )

        else:
            message = _("{nick} has left{reason}").format(nick=nick, reason=reason)
        return message

    @staticmethod
    def _make_join_message(nick: str) -> str:
        return _("%s has joined") % nick



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# Chat Markers (XEP-0333)

from __future__ import annotations

from typing import Any

from dataclasses import dataclass
from datetime import datetime

import nbxmpp
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler

from gajim.common import app
from gajim.common import types
from gajim.common.events import DisplayedReceived
from gajim.common.events import ReadStateSync
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.modules.contacts import ResourceContact
from gajim.common.modules.message_util import get_chat_type_and_direction
from gajim.common.modules.message_util import get_message_timestamp
from gajim.common.modules.message_util import get_occupant_info
from gajim.common.storage.archive import models as mod
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageType
from gajim.common.structs import OutgoingMessage


class ChatMarkers(BaseModule):

    _nbxmpp_extends = 'ChatMarkers'

    def __init__(self, client: types.Client):
        BaseModule.__init__(self, client)

        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_chat_marker,
                          ns=Namespace.CHATMARKERS,
                          priority=47),
        ]

    def pass_disco(self, info: DiscoInfo) -> None:
        self.mds_assist_supported = Namespace.MDS_ASSIST in info.features

    def _process_chat_marker(self,
                             _client: types.NBXMPPClient,
                             _stanza: Any,
                             properties: MessageProperties) -> None:

        if properties.marker is None or not properties.marker.is_displayed:
            return

        if properties.type.is_error:
            return

        self._process(properties)
        raise nbxmpp.NodeProcessed

    def _process(self, properties: MessageProperties) -> None:
        assert properties.marker is not None
        assert properties.jid is not None

        jid = properties.jid
        if not properties.is_muc_pm:
            jid = properties.jid.new_as_bare()

        if properties.type.is_groupchat:
            if properties.jid.resource is None:
                return

            assert properties.muc_jid is not None
            contact = self._client.get_module('Contacts').get_contact(
                properties.muc_jid,
                groupchat=True)
            assert isinstance(contact, GroupchatContact)
            if not contact.is_joined:
                self._log.warning('Received chat marker while not joined')
                return

            if properties.muc_nickname != contact.nickname:
                self._raise_event(properties)
                return

            self._raise_read_state_sync(jid, properties.marker.id)
            return

        if (properties.is_sent_carbon or
                (properties.is_mam_message and properties.is_from_us())):
            self._raise_read_state_sync(jid, properties.marker.id)
            return

        self._raise_event(properties)

    def _raise_read_state_sync(self, jid: JID, marker_id: str) -> None:
        self._log.info('Read state sync: %s - %s', jid, marker_id)
        app.ged.raise_event(
            ReadStateSync(account=self._account,
                          jid=jid,
                          marker_id=marker_id))

    def _raise_event(self, properties: MessageProperties) -> None:
        assert properties.marker is not None
        assert properties.jid is not None
        assert properties.remote_jid is not None

        self._log.info('displayed-received: %s %s',
                       properties.remote_jid,
                       properties.marker.id)

        remote_jid = properties.remote_jid
        timestamp = get_message_timestamp(properties)

        muc_data = None
        if properties.type.is_groupchat:
            muc_data = self._client.get_module('MUC').get_muc_data(remote_jid)
            if muc_data is None:
                self._log.warning(
                    'Groupchat message from unknown MUC: %s', remote_jid
                )
                return

        m_type, direction = get_chat_type_and_direction(
            muc_data, self._client.get_own_jid(), properties)

        if direction == ChatDirection.OUTGOING:
            return

        contact = self._get_contact_with_mtype(m_type, properties.jid)

        if not self._is_sending_marker_allowed(contact):
            self._log.info('Ignore marker because setting is disabled')
            return

        occupant = None
        if m_type in (MessageType.GROUPCHAT, MessageType.PM):
            if properties.jid.is_bare:
                self._log.warning("Received marker from MUC bare jid")
                return

            assert isinstance(contact, GroupchatParticipant)
            occupant = get_occupant_info(
                self._account,
                remote_jid=remote_jid,
                own_bare_jid=self._get_own_bare_jid(),
                direction=ChatDirection.INCOMING,
                m_type=m_type,
                timestamp=timestamp,
                contact=contact,
                properties=properties,
            )

            if occupant is None:
                # Support chat markers in group chats only if occupant-id
                # is available
                return

        marker_data = mod.DisplayedMarker(
            account_=self._account,
            remote_jid_=properties.remote_jid,
            occupant_=occupant,
            id=properties.marker.id,
            timestamp=timestamp)

        pk = app.storage.archive.insert_object(marker_data)
        if pk == -1:
            return

        app.ged.raise_event(
            DisplayedReceived(
                account=self._account,
                jid=properties.remote_jid,
                pk=pk,
            )
        )

    def send_displayed_marker(self,
                              contact: types.ChatContactT,
                              message_id: str,
                              stanza_id: str | None) -> bool:

        # Return value is True if displayed marker was sent and
        # mds assist was added

        if not self._is_sending_marker_allowed(contact):
            return False

        marker_id = self._determine_marker_id(contact, message_id, stanza_id)

        if not self.mds_assist_supported or isinstance(contact, ResourceContact):
            # Assist is not added for private groupchat messages see
            # https://xmpp.org/extensions/xep-0490.html#rules-client
            stanza_id = None

        message = OutgoingMessage(account=self._account,
                                  contact=contact,
                                  marker=('displayed', marker_id),
                                  mds_id=stanza_id,
                                  play_sound=False)

        self._client.send_message(message)
        self._log.info('Send displayed to %s, marker id: %s, mds id: %s',
                        contact.jid, marker_id, stanza_id)

        return stanza_id is not None

    @staticmethod
    def _determine_marker_id(contact: types.ChatContactT,
                             message_id: str,
                             stanza_id: str | None
                             ) -> str:

        if stanza_id is None:
            return message_id

        if isinstance(contact, GroupchatContact):
            return stanza_id
        return message_id

    def _get_contact_with_mtype(
        self, mtype: MessageType,
        jid: JID
    ) -> types.ChatContactT:

        if mtype in (MessageType.GROUPCHAT, MessageType.PM):
            contact = self._get_contact(jid, groupchat=True)
        else:
            contact = self._get_contact(jid.new_as_bare(), groupchat=False)
        assert not isinstance(contact, ResourceContact)
        return contact

    @staticmethod
    def _is_sending_marker_allowed(contact: types.ChatContactT) -> bool:
        if isinstance(contact, BareContact):
            if not contact.is_subscribed:
                return False

            return app.settings.get_contact_setting(
                contact.account, contact.jid, 'send_marker')

        return app.settings.get_group_chat_setting(
            contact.account, contact.jid.new_as_bare(), 'send_marker')


@dataclass
class DisplayedMarkerData:
    account: str
    jid: JID
    id: str
    timestamp: datetime
    occupant: mod.Occupant | None

    @classmethod
    def from_model(
        cls,
        account: str,
        marker: mod.DisplayedMarker
    ) -> DisplayedMarkerData:
        return cls(
            account=account,
            jid=marker.remote.jid,
            id=marker.id,
            timestamp=marker.timestamp,
            occupant=marker.occupant
        )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import logging

from gi.repository import Gtk
from nbxmpp.errors import StanzaError
from nbxmpp.task import Task

from gajim.common import app
from gajim.common.client import Client
from gajim.common.i18n import _
from gajim.common.modules.contacts import GroupchatContact

from gajim.gtk.apply_button_box import ApplyButtonBox
from gajim.gtk.builder import get_builder
from gajim.gtk.dataform import DataFormWidget
from gajim.gtk.util.classes import SignalManager

log = logging.getLogger("gajim.gtk.groupchat_config")


class GroupchatConfig(Gtk.Box, SignalManager):
    def __init__(self, client: Client, contact: GroupchatContact) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        self._contact = contact
        self._client = client
        self._data_form_widget: DataFormWidget | None = None
        self_contact = self._contact.get_self()
        assert self_contact is not None

        self._own_affiliation = self_contact.affiliation

        self._ui = get_builder("groupchat_config.ui")

        self._apply_button = ApplyButtonBox(_("Apply"), self._on_apply)
        self._apply_button.set_halign(Gtk.Align.END)
        self._apply_button.add_css_class("m-18")

        self._ui.config_box.append(self._apply_button)

        self.append(self._ui.stack)

        if self._own_affiliation.is_owner:
            self._ui.stack.set_visible_child_name("loading")
            self._client.get_module("MUC").request_config(
                self._contact.jid, callback=self._on_config_received
            )
        else:
            self._ui.error_label.set_text(
                _("You need Owner permission to change the configuration")
            )
            self._ui.error_image.set_from_icon_name("lucide-circle-x-symbolic")
            self._ui.stack.set_visible_child_name("error")

    def do_unroot(self) -> None:
        self._disconnect_all()
        del self._apply_button
        if self._data_form_widget is not None:
            app.check_finalize(self._data_form_widget)
        del self._data_form_widget
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)

    def _set_form(self, form: Any) -> None:
        self._data_form_widget = DataFormWidget(form)
        self._connect(self._data_form_widget, "is-valid", self._on_is_valid)
        self._data_form_widget.add_css_class("p-12")
        self._ui.config_box.prepend(self._data_form_widget)
        self._ui.stack.set_visible_child_name("config")

    def _on_is_valid(self, _widget: Any, is_valid: bool) -> None:
        assert self._data_form_widget is not None
        if not self._data_form_widget.was_modified():
            self._apply_button.set_button_state(False)
        else:
            self._apply_button.set_button_state(is_valid)

    def _on_config_received(self, task: Task) -> None:
        try:
            result = task.finish()
        except StanzaError as error:
            self._ui.error_label.set_text(str(error))
            self._ui.error_image.set_from_icon_name("lucide-circle-x-symbolic")
            self._ui.stack.set_visible_child_name("error")
            return

        self._set_form(result.form)
        self._ui.stack.set_visible_child_name("config")

    def _on_apply(self, button: Gtk.Button) -> None:
        assert self._data_form_widget is not None
        form = self._data_form_widget.get_submit_form()

        self._data_form_widget.set_sensitive(False)

        self._client.get_module("MUC").set_config(
            self._contact.jid, form, callback=self._on_finished
        )

    def _on_finished(self, task: Task) -> None:
        assert self._data_form_widget is not None
        self._data_form_widget.set_sensitive(True)
        self._data_form_widget.reset_form_hash()

        try:
            task.finish()
        except StanzaError as error:
            self._apply_button.set_error(str(error))
            return

        self._apply_button.set_success()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0012: Last Activity

from __future__ import annotations

from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common import idle
from gajim.common import types
from gajim.common.modules.base import BaseModule


class LastActivity(BaseModule):

    _nbxmpp_extends = 'LastActivity'
    _nbxmpp_methods = [
        'request_last_activity',
        'set_idle_func',
        'disable',
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

    def set_enabled(self, enabled: bool) -> None:
        if not enabled or not app.is_installed('IDLE'):
            self._nbxmpp('LastActivity').disable()
            return

        if not app.settings.get_account_setting(self._account,
                                                'send_idle_time'):
            return

        self._nbxmpp('LastActivity').set_idle_func(idle.Monitor.get_idle_sec)
        self._nbxmpp('LastActivity').set_allow_reply_func(self._allow_reply)

    def _allow_reply(self, jid: JID) -> bool:
        bare_jid = jid.new_as_bare()
        item = self._con.get_module('Roster').get_item(bare_jid)
        if item is None:
            return False

        return item.subscription in ('from', 'both')
